"""Classes and functions for fitting the covariance tensor model of q-space
trajectory imaging (QTI) by :footcite:t:`Westin2016`.

References
----------
.. footbibliography::
"""

from warnings import warn

import numpy as np

from dipy.reconst.base import ReconstModel
from dipy.reconst.dti import auto_attr
from dipy.testing.decorators import warning_for_keywords
from dipy.utils.optpkg import optional_package

cp, have_cvxpy, _ = optional_package("cvxpy", min_version="1.4.1")


# XXX Eventually to be replaced with `reconst.dti.lower_triangular`
def from_3x3_to_6x1(T):
    """Convert symmetric 3 x 3 matrices into 6 x 1 vectors.

    Parameters
    ----------
    T : numpy.ndarray
        An array of size (..., 3, 3).

    Returns
    -------
    V : numpy.ndarray
        Converted vectors of size (..., 6, 1).

    Notes
    -----
    The conversion of a matrix into a vector is defined as

        .. math::

            \\mathbf{V} = \\begin{bmatrix}
            T_{11} & T_{22} & T_{33} &
            \\sqrt{2} T_{23} & \\sqrt{2} T_{13} & \\sqrt{2} T_{12}
            \\end{bmatrix}^T
    """
    if T.shape[-2::] != (3, 3):
        raise ValueError("The shape of the input array must be (..., 3, 3).")
    if not np.all(np.isclose(T, np.swapaxes(T, -1, -2))):
        warn(
            "All matrices converted to Voigt notation are not symmetric.", stacklevel=2
        )
    C = np.sqrt(2)
    V = np.stack(
        (
            T[..., 0, 0],
            T[..., 1, 1],
            T[..., 2, 2],
            C * T[..., 1, 2],
            C * T[..., 0, 2],
            C * T[..., 0, 1],
        ),
        axis=-1,
    )[..., np.newaxis]
    return V


def from_6x1_to_3x3(V):
    """Convert 6 x 1 vectors into symmetric 3 x 3 matrices.

    Parameters
    ----------
    V : numpy.ndarray
        An array of size (..., 6, 1).

    Returns
    -------
    T : numpy.ndarray
        Converted matrices of size (..., 3, 3).

    Notes
    -----
    The conversion of a matrix into a vector is defined as

        .. math::

            \\mathbf{V} = \\begin{bmatrix}
            T_{11} & T_{22} & T_{33} &
            \\sqrt{2} T_{23} & \\sqrt{2} T_{13} & \\sqrt{2} T_{12}
            \\end{bmatrix}^T
    """
    if V.shape[-2::] != (6, 1):
        raise ValueError("The shape of the input array must be (..., 6, 1).")
    C = 1 / np.sqrt(2)
    T = np.array(
        (
            [V[..., 0, 0], C * V[..., 5, 0], C * V[..., 4, 0]],
            [C * V[..., 5, 0], V[..., 1, 0], C * V[..., 3, 0]],
            [C * V[..., 4, 0], C * V[..., 3, 0], V[..., 2, 0]],
        )
    )
    T = np.moveaxis(T, (0, 1), (-2, -1))
    return T


def from_6x6_to_21x1(T):
    """Convert symmetric 6 x 6 matrices into 21 x 1 vectors.

    Parameters
    ----------
    T : numpy.ndarray
        An array of size (..., 6, 6).

    Returns
    -------
    V : numpy.ndarray
        Converted vectors of size (..., 21, 1).

    Notes
    -----
    The conversion of a matrix into a vector is defined as

        .. math::

            \\begin{matrix}
            \\mathbf{V} = & \\big[
            T_{11} & T_{22} & T_{33} \\
            & \\sqrt{2} T_{23} & \\sqrt{2} T_{13} & \\sqrt{2} T_{12} \\
            & \\sqrt{2} T_{14} & \\sqrt{2} T_{15} & \\sqrt{2} T_{16} \\
            & \\sqrt{2} T_{24} & \\sqrt{2} T_{25} & \\sqrt{2} T_{26} \\
            & \\sqrt{2} T_{34} & \\sqrt{2} T_{35} & \\sqrt{2} T_{36} \\
            & T_{44} & T_{55} & T_{66} \\
            & \\sqrt{2} T_{45} & \\sqrt{2} T_{56} & \\sqrt{2} T_{46} \\big]^T
            \\end{matrix}
    """
    if T.shape[-2::] != (6, 6):
        raise ValueError("The shape of the input array must be (..., 6, 6).")
    if not np.all(np.isclose(T, np.swapaxes(T, -1, -2), equal_nan=True)):
        warn(
            "All matrices converted to Voigt notation are not symmetric.", stacklevel=2
        )
    C = np.sqrt(2)
    V = np.stack(
        (
            [
                T[..., 0, 0],
                T[..., 1, 1],
                T[..., 2, 2],
                C * T[..., 1, 2],
                C * T[..., 0, 2],
                C * T[..., 0, 1],
                C * T[..., 0, 3],
                C * T[..., 0, 4],
                C * T[..., 0, 5],
                C * T[..., 1, 3],
                C * T[..., 1, 4],
                C * T[..., 1, 5],
                C * T[..., 2, 3],
                C * T[..., 2, 4],
                C * T[..., 2, 5],
                T[..., 3, 3],
                T[..., 4, 4],
                T[..., 5, 5],
                C * T[..., 3, 4],
                C * T[..., 4, 5],
                C * T[..., 3, 5],
            ]
        ),
        axis=-1,
    )[..., np.newaxis]
    return V


def from_21x1_to_6x6(V):
    """Convert 21 x 1 vectors into symmetric 6 x 6 matrices.

    Parameters
    ----------
    V : numpy.ndarray
        An array of size (..., 21, 1).

    Returns
    -------
    T : numpy.ndarray
        Converted matrices of size (..., 6, 6).

    Notes
    -----
    The conversion of a matrix into a vector is defined as

        .. math::

            \\begin{matrix}
            \\mathbf{V} = & \\big[
            T_{11} & T_{22} & T_{33} \\
            & \\sqrt{2} T_{23} & \\sqrt{2} T_{13} & \\sqrt{2} T_{12} \\
            & \\sqrt{2} T_{14} & \\sqrt{2} T_{15} & \\sqrt{2} T_{16} \\
            & \\sqrt{2} T_{24} & \\sqrt{2} T_{25} & \\sqrt{2} T_{26} \\
            & \\sqrt{2} T_{34} & \\sqrt{2} T_{35} & \\sqrt{2} T_{36} \\
            & T_{44} & T_{55} & T_{66} \\
            & \\sqrt{2} T_{45} & \\sqrt{2} T_{56} & \\sqrt{2} T_{46} \\big]^T
            \\end{matrix}
    """
    if V.shape[-2::] != (21, 1):
        raise ValueError("The shape of the input array must be (..., 21, 1).")
    C = 1 / np.sqrt(2)
    T = np.array(
        (
            [
                V[..., 0, 0],
                C * V[..., 5, 0],
                C * V[..., 4, 0],
                C * V[..., 6, 0],
                C * V[..., 7, 0],
                C * V[..., 8, 0],
            ],
            [
                C * V[..., 5, 0],
                V[..., 1, 0],
                C * V[..., 3, 0],
                C * V[..., 9, 0],
                C * V[..., 10, 0],
                C * V[..., 11, 0],
            ],
            [
                C * V[..., 4, 0],
                C * V[..., 3, 0],
                V[..., 2, 0],
                C * V[..., 12, 0],
                C * V[..., 13, 0],
                C * V[..., 14, 0],
            ],
            [
                C * V[..., 6, 0],
                C * V[..., 9, 0],
                C * V[..., 12, 0],
                V[..., 15, 0],
                C * V[..., 18, 0],
                C * V[..., 20, 0],
            ],
            [
                C * V[..., 7, 0],
                C * V[..., 10, 0],
                C * V[..., 13, 0],
                C * V[..., 18, 0],
                V[..., 16, 0],
                C * V[..., 19, 0],
            ],
            [
                C * V[..., 8, 0],
                C * V[..., 11, 0],
                C * V[..., 14, 0],
                C * V[..., 20, 0],
                C * V[..., 19, 0],
                V[..., 17, 0],
            ],
        )
    )
    T = np.moveaxis(T, (0, 1), (-2, -1))
    return T


def cvxpy_1x6_to_3x3(V):
    """Convert a 1 x 6 vector into a symmetric 3 x 3 matrix.

    Parameters
    ----------
    V : numpy.ndarray
        An array of size (1, 6).

    Returns
    -------
    T : cvxpy.bmat
        Converted matrix of size (3, 3).

    Notes
    -----
    The conversion of a matrix into a vector is defined as

        .. math::

            \\mathbf{V} = \\begin{bmatrix}
            T_{11} & T_{22} & T_{33} &
            \\sqrt{2} T_{23} & \\sqrt{2} T_{13} & \\sqrt{2} T_{12}
            \\end{bmatrix}^T
    """
    if V.shape[0] == 6:
        V = V.T

    f = 1 / np.sqrt(2)

    T = cp.bmat(
        [
            [V[0, 0], f * V[0, 5], f * V[0, 4]],
            [f * V[0, 5], V[0, 1], f * V[0, 3]],
            [f * V[0, 4], f * V[0, 3], V[0, 2]],
        ]
    )
    return T


def cvxpy_1x21_to_6x6(V):
    """Convert 1 x 21 vector into a symmetric 6 x 6 matrix.

    Parameters
    ----------
    V : numpy.ndarray
        An array of size (1, 21).

    Returns
    -------
    T : cvxpy.bmat
        Converted matrices of size (6, 6).

    Notes
    -----
    The conversion of a matrix into a vector is defined as

        .. math::

            \\begin{matrix}
            \\mathbf{V} = & \\big[
            T_{11} & T_{22} & T_{33} \\
            & \\sqrt{2} T_{23} & \\sqrt{2} T_{13} & \\sqrt{2} T_{12} \\
            & \\sqrt{2} T_{14} & \\sqrt{2} T_{15} & \\sqrt{2} T_{16} \\
            & \\sqrt{2} T_{24} & \\sqrt{2} T_{25} & \\sqrt{2} T_{26} \\
            & \\sqrt{2} T_{34} & \\sqrt{2} T_{35} & \\sqrt{2} T_{36} \\
            & T_{44} & T_{55} & T_{66} \\
            & \\sqrt{2} T_{45} & \\sqrt{2} T_{56} & \\sqrt{2} T_{46} \\big]^T
            \\end{matrix}
    """
    if V.shape[0] == 21:
        V = V.T

    f = 1 / np.sqrt(2)

    T = cp.bmat(
        [
            [V[0, 0], f * V[0, 5], f * V[0, 4], f * V[0, 6], f * V[0, 7], f * V[0, 8]],
            [
                f * V[0, 5],
                V[0, 1],
                f * V[0, 3],
                f * V[0, 9],
                f * V[0, 10],
                f * V[0, 11],
            ],
            [
                f * V[0, 4],
                f * V[0, 3],
                V[0, 2],
                f * V[0, 12],
                f * V[0, 13],
                f * V[0, 14],
            ],
            [
                f * V[0, 6],
                f * V[0, 9],
                f * V[0, 12],
                V[0, 15],
                f * V[0, 18],
                f * V[0, 20],
            ],
            [
                f * V[0, 7],
                f * V[0, 10],
                f * V[0, 13],
                f * V[0, 18],
                V[0, 16],
                f * V[0, 19],
            ],
            [
                f * V[0, 8],
                f * V[0, 11],
                f * V[0, 14],
                f * V[0, 20],
                f * V[0, 19],
                V[0, 17],
            ],
        ]
    )
    return T


# These tensors are used in the calculation of the QTI parameters
e_iso = np.eye(3) / 3
E_iso = np.eye(6) / 3
E_bulk = from_3x3_to_6x1(e_iso) @ from_3x3_to_6x1(e_iso).T
E_shear = E_iso - E_bulk
E_tsym = E_bulk + 0.4 * E_shear


def dtd_covariance(DTD):
    """Calculate covariance of a diffusion tensor distribution (DTD).

    Parameters
    ----------
    DTD : numpy.ndarray
        Diffusion tensor distribution of shape (number of tensors, 3, 3) or
        (number of tensors, 6, 1).

    Returns
    -------
    C : numpy.ndarray
        Covariance tensor of shape (6, 6).

    Notes
    -----
    The covariance tensor is calculated according to the following equation and
    converted into a rank-2 tensor :footcite:p:`Westin2016`:

        .. math::

            \\mathbb{C} = \\langle \\mathbf{D} \\otimes \\mathbf{D} \\rangle -
            \\langle \\mathbf{D} \\rangle \\otimes \\langle \\mathbf{D}
            \\rangle

    References
    ----------
    .. footbibliography::
    """
    dims = DTD.shape
    if len(dims) != 3 or (dims[1:3] != (3, 3) and dims[1:3] != (6, 1)):
        raise ValueError(
            "The shape of DTD must be (number of tensors, 3, 3) or (number of "
            + "tensors, 6, 1)."
        )
    if dims[1:3] == (3, 3):
        DTD = from_3x3_to_6x1(DTD)
    D = np.mean(DTD, axis=0)
    C = np.mean(DTD @ np.swapaxes(DTD, -2, -1), axis=0) - D @ np.swapaxes(D, -2, -1)
    return C


@warning_for_keywords()
def qti_signal(gtab, D, C, *, S0=1):
    """Generate signals using the covariance tensor signal representation.

    Parameters
    ----------
    gtab : dipy.core.gradients.GradientTable
        Gradient table with b-tensors.
    D : numpy.ndarray
        Diffusion tensors of shape (..., 3, 3), (..., 6, 1), or (..., 6).
    C : numpy.ndarray
        Covariance tensors of shape (..., 6, 6), (..., 21, 1), or (..., 21).
    S0 : numpy.ndarray, optional
        Signal magnitudes without diffusion-weighting. Must be a single number
        or an array of same shape as D and C without the last two dimensions.

    Returns
    -------
    S : numpy.ndarray
        Simulated signals.

    Notes
    -----
    The signal is generated according to

        .. math::

            S = S_0 \\exp \\left(- \\mathbf{b} : \\langle \\mathbf{D} \\rangle
            + \\frac{1}{2}(\\mathbf{b} \\otimes \\mathbf{b}) : \\mathbb{C}
            \\right)
    """

    # Validate input and convert to Voigt notation if necessary
    if gtab.btens is None:
        raise ValueError("QTI requires b-tensors to be defined in the gradient table.")
    if D.shape[-2::] != (6, 1):
        if D.shape[-2::] == (3, 3):
            D = from_3x3_to_6x1(D)
        elif D.shape[-1] == 6:
            D = D[..., np.newaxis]
        else:
            raise ValueError(
                "The shape of D must be (..., 3, 3), (..., 6, 1) or (..., 6)."
            )
    if C.shape[-2::] != (21, 1):
        if C.shape[-2::] == (6, 6):
            C = from_6x6_to_21x1(C)
        elif C.shape[-1] == 21:
            C = C[..., np.newaxis]
        else:
            raise ValueError(
                "The shape of C must be (..., 6, 6), (..., 21, 1), or (..., 21)."
            )
    if D.shape[0:-2] != C.shape[0:-2]:
        raise ValueError("The shapes of C and D are not compatible")
    if not isinstance(S0, (int, float)):
        if S0.shape != (1,) and S0.shape != D.shape[0:-2]:
            raise ValueError(
                "S0 must be a single number or an array of the same shape "
                + " compatible with D and C."
            )

    # Generate signals
    S = np.zeros(D.shape[0:-2] + (gtab.btens.shape[0],))
    for i, bten in enumerate(gtab.btens):
        b = from_3x3_to_6x1(bten)
        b_sq = from_6x6_to_21x1(b @ np.swapaxes(b, -2, -1))
        S[..., i] = (
            S0
            * np.exp(-np.swapaxes(b, -2, -1) @ D + 0.5 * np.swapaxes(b_sq, -2, -1) @ C)[
                ..., 0, 0
            ]
        )
    return S


def design_matrix(btens):
    """Calculate the design matrix from the b-tensors.

    Parameters
    ----------
    btens : numpy.ndarray
        An array of b-tensors of shape (number of acquisitions, 3, 3).

    Returns
    -------
    X : numpy.ndarray
        Design matrix.

    Notes
    -----
    The design matrix is generated according to

        .. math::

            X = \\begin{pmatrix} 1 & -\\mathbf{b}_1^T & \\frac{1}{2}(
            \\mathbf{b}_1
            \\otimes\\mathbf{b}_1)^T \\ \\vdots & \\vdots & \\vdots \\ 1 &
            -\\mathbf{b}_n^T & \\frac{1}{2}(\\mathbf{b}_n\\otimes
            \\mathbf{b}_n)^T
            \\end{pmatrix}
    """
    X = np.zeros((btens.shape[0], 28))
    for i, bten in enumerate(btens):
        b = from_3x3_to_6x1(bten)
        b_sq = from_6x6_to_21x1(b @ b.T)
        X[i] = np.concatenate(([1], (-b.T)[0, :], (0.5 * b_sq.T)[0, :]))
    return X


@warning_for_keywords()
def _ols_fit(data, mask, X, *, step=int(1e4)):
    """Estimate the model parameters using ordinary least squares.

    Parameters
    ----------
    data : numpy.ndarray
        Array of shape (..., number of acquisitions).
    mask : numpy.ndarray
        Boolean array with the same shape as the data array of a single
        acquisition.
    X : numpy.ndarray
        Design matrix of shape (number of acquisitions, 28).
    step : int, optional
        The number of voxels over which the fit is calculated simultaneously.

    Returns
    -------
    params : numpy.ndarray
        Array of shape (..., 28) containing the estimated model parameters.
        Element 0 is the natural logarithm of the estimated signal without
        diffusion-weighting, elements 1-6 are the estimated diffusion tensor
        elements in Voigt notation, and elements 7-27 are the estimated
        covariance tensor elements in Voigt notation.
    """
    params = np.zeros((np.prod(mask.shape), 28)) * np.nan
    data_masked = data[mask]
    size = len(data_masked)
    X_inv = np.linalg.pinv(X.T @ X)  # Independent of data
    if step >= size:  # Fit over all data simultaneously
        S = np.log(data_masked)[..., np.newaxis]
        params_masked = (X_inv @ X.T @ S)[..., 0]
    else:  # Iterate over data
        params_masked = np.zeros((size, 28))
        for i in range(0, size, step):
            S = np.log(data_masked[i : i + step])[..., np.newaxis]
            params_masked[i : i + step] = (X_inv @ X.T @ S)[..., 0]
    params[np.where(mask.ravel())] = params_masked
    params = params.reshape((mask.shape + (28,)))
    return params


@warning_for_keywords()
def _wls_fit(data, mask, X, *, step=int(1e4)):
    """Estimate the model parameters using weighted least squares with the
    signal magnitudes as weights.

    Parameters
    ----------
    data : numpy.ndarray
        Array of shape (..., number of acquisitions).
    mask : numpy.ndarray
        Array with the same shape as the data array of a single acquisition.
    X : numpy.ndarray
        Design matrix of shape (number of acquisitions, 28).
    step : int, optional
        The number of voxels over which the fit is calculated simultaneously.

    Returns
    -------
    params : numpy.ndarray
        Array of shape (..., 28) containing the estimated model parameters.
        Element 0 is the natural logarithm of the estimated signal without
        diffusion-weighting, elements 1-6 are the estimated diffusion tensor
        elements in Voigt notation, and elements 7-27 are the estimated
        covariance tensor elements in Voigt notation.
    """
    params = np.zeros((np.prod(mask.shape), 28)) * np.nan
    data_masked = data[mask]
    size = len(data_masked)
    if step >= size:  # Fit over all data simultaneously
        S = np.log(data_masked)[..., np.newaxis]
        C = data_masked[:, np.newaxis, :]
        B = X.T * C
        A = np.linalg.pinv(B @ X)
        params_masked = (A @ B @ S)[..., 0]
    else:  # Iterate over data
        params_masked = np.zeros((size, 28))
        for i in range(0, size, step):
            S = np.log(data_masked[i : i + step])[..., np.newaxis]
            C = data_masked[i : i + step][:, np.newaxis, :]
            B = X.T * C
            A = np.linalg.pinv(B @ X)
            params_masked[i : i + step] = (A @ B @ S)[..., 0]
    params[np.where(mask.ravel())] = params_masked
    params = params.reshape((mask.shape + (28,)))
    return params


def _sdpdc_fit(data, mask, X, cvxpy_solver):
    """Estimate the model parameters using Semidefinite Programming (SDP),
    while enforcing positivity constraints on the D and C tensors (SDPdc).

    See :footcite:p:`Herberthson2021` for further details about the method.

    Parameters
    ----------
    data : numpy.ndarray
        Array of shape (..., number of acquisitions).
    mask : numpy.ndarray
        Array with the same shape as the data array of a single acquisition.
    X : numpy.ndarray
        Design matrix of shape (number of acquisitions, 28).
    cvxpy_solver: string, required
        The name of the SDP solver to be used. Default: 'SCS'

    Returns
    -------
    params : numpy.ndarray
        Array of shape (..., 28) containing the estimated model parameters.
        Element 0 is the natural logarithm of the estimated signal without
        diffusion-weighting, elements 1-6 are the estimated diffusion tensor
        elements in Voigt notation, and elements 7-27 are the estimated
        covariance tensor elements in Voigt notation.

    References
    ----------
    .. footbibliography::
    """

    if not have_cvxpy:
        raise ImportError("CVXPY package needed to enforce constraints")

    if cvxpy_solver not in cp.installed_solvers():
        raise ValueError("The selected solver is not available")

    params = np.zeros((np.prod(mask.shape), 28)) * np.nan
    data_masked = data[mask]
    size, nvols = data_masked.shape
    scale = np.maximum(np.max(data_masked, axis=1, keepdims=True), 1)
    data_masked = data_masked / scale
    data_masked[data_masked < 0] = 0
    log_data = np.log(data_masked)
    params_masked = np.zeros((size, 28))

    x = cp.Variable((28, 1))
    y = cp.Parameter((nvols, 1))
    A = cp.Parameter((nvols, 28))
    dc = cvxpy_1x6_to_3x3(x[1:7])
    cc = cvxpy_1x21_to_6x6(x[7:])
    constraints = [dc >> 0, cc >> 0]
    objective = cp.Minimize(cp.norm(A @ x - y))
    prob = cp.Problem(objective, constraints)
    unconstrained = cp.Problem(objective)

    for i in range(0, size, 1):
        vox_data = data_masked[i : i + 1, :].T
        vox_log_data = log_data[i : i + 1, :].T
        vox_log_data[np.isinf(vox_log_data)] = 0
        y.value = vox_data * vox_log_data
        A.value = vox_data * X

        try:
            prob.solve(solver=cvxpy_solver, verbose=False)
            m = x.value
        except Exception:
            msg = "Constrained optimization failed, attempting unconstrained"
            msg += " optimization."
            warn(msg, stacklevel=2)
            try:
                unconstrained.solve(solver=cvxpy_solver)
                m = x.value
            except Exception:
                msg = "Unconstrained optimization failed,"
                msg += " returning zero array."
                warn(msg, stacklevel=2)
                m = np.zeros(x.shape)

        params_masked[i : i + 1, :] = m.T

    params_masked[:, 0] += np.log(scale[:, 0])
    params[np.where(mask.ravel())] = params_masked
    params = params.reshape((mask.shape + (28,)))
    return params


class QtiModel(ReconstModel):
    @warning_for_keywords()
    def __init__(self, gtab, *, fit_method="WLS", cvxpy_solver="SCS"):
        """Covariance tensor model of q-space trajectory imaging.

        See :footcite:t:`Westin2016` for further details about the model.

        Parameters
        ----------
        gtab : dipy.core.gradients.GradientTable
            Gradient table with b-tensors.
        fit_method : str, optional
            Must be one of the following:

            - 'OLS' for ordinary least squares :func:`qti._ols_fit`
            - 'WLS' for weighted least squares :func:`qti._wls_fit`
            - 'SDPDc' for semidefinite programming with positivity constraints
              applied :footcite:p:`Herberthson2021` :func:`qti._sdpdc_fit`

        cvxpy_solver: str, optionals
            solver for the SDP formulation. default: 'SCS'

        References
        ----------
        .. footbibliography::
        """
        ReconstModel.__init__(self, gtab)

        if self.gtab.btens is None:
            raise ValueError(
                "QTI requires b-tensors to be defined in the gradient table."
            )
        self.X = design_matrix(self.gtab.btens)
        rank = np.linalg.matrix_rank(self.X.T @ self.X)
        if rank < 28:
            warn(
                "The combination of the b-tensor shapes, sizes, and "
                + "orientations does not enable all elements of the covariance "
                + f"tensor to be estimated (rank(X.T @ X) = {rank} < 28).",
                stacklevel=2,
            )

        try:
            self.fit_method = common_fit_methods[fit_method]
        except KeyError as e:
            raise ValueError(
                f"Invalid value ({fit_method}) for 'fit_method'."
                + " Options: 'OLS', 'WLS', 'SDPdc'."
            ) from e

        self.cvxpy_solver = cvxpy_solver
        self.fit_method_name = fit_method

    @warning_for_keywords()
    def fit(self, data, *, mask=None):
        """Fit QTI to data.

        Parameters
        ----------
        data : numpy.ndarray
            Array of shape (..., number of acquisitions).
        mask : numpy.ndarray, optional
            Array with the same shape as the data array of a single acquisition.

        Returns
        -------
        qtifit : dipy.reconst.qti.QtiFit
            The fitted model.
        """
        if mask is None:
            mask = np.ones(data.shape[:-1], dtype=bool)
        else:
            if mask.shape != data.shape[:-1]:
                raise ValueError("Mask is not the same shape as data.")
            mask = np.asarray(mask, dtype=bool)
        if self.fit_method_name == "SDPdc":
            params = self.fit_method(data, mask, self.X, self.cvxpy_solver)
        else:
            params = self.fit_method(data, mask, self.X)
        return QtiFit(params)

    def predict(self, params):
        """Generate signals from this model class instance and given parameters.

        Parameters
        ----------
        params : numpy.ndarray
            Array of shape (..., 28) containing the model parameters. Element 0
            is the natural logarithm of the signal without diffusion-weighting,
            elements 1-6 are the diffusion tensor elements in Voigt notation,
            and elements 7-27 are the covariance tensor elements in Voigt
            notation.

        Returns
        -------
        S : numpy.ndarray
            Signals.
        """
        S0 = np.exp(params[..., 0])
        D = params[..., 1:7, np.newaxis]
        C = params[..., 7::, np.newaxis]
        S = qti_signal(self.gtab, D, C, S0=S0)
        return S


class QtiFit:
    def __init__(self, params):
        """Fitted QTI model.

        Parameters
        ----------
        params : numpy.ndarray
            Array of shape (..., 28) containing the model parameters. Element 0
            is the natural logarithm of the signal without diffusion-weighting,
            elements 1-6 are the diffusion tensor elements in Voigt notation,
            and elements 7-27 are the covariance tensor elements in Voigt
            notation.
        """
        self.params = params

    def predict(self, gtab):
        """Generate signals from this model fit and a given gradient table.

        Parameters
        ----------
        gtab : dipy.core.gradients.GradientTable
            Gradient table with b-tensors.

        Returns
        -------
        S : numpy.ndarray
            Signals.
        """
        if gtab.btens is None:
            raise ValueError(
                "QTI requires b-tensors to be defined in the gradient table."
            )
        S0 = self.S0_hat
        D = self.params[..., 1:7, np.newaxis]
        C = self.params[..., 7::, np.newaxis]
        S = qti_signal(gtab, D, C, S0=S0)
        return S

    @auto_attr
    def S0_hat(self):
        """Estimated signal without diffusion-weighting.

        Returns
        -------
        S0 : numpy.ndarray
        """
        S0 = np.exp(self.params[..., 0])
        return S0

    @auto_attr
    def md(self):
        """Mean diffusivity.

        Returns
        -------
        md : numpy.ndarray

        Notes
        -----
        Mean diffusivity is calculated as

            .. math::

                \\text{MD} = \\langle \\mathbf{D} \\rangle :
                \\mathbf{E}_\\text{iso}
        """
        md = np.matmul(self.params[..., np.newaxis, 1:7], from_3x3_to_6x1(e_iso))[
            ..., 0, 0
        ]
        return md

    @auto_attr
    def v_md(self):
        """Variance of microscopic mean diffusivities.

        Returns
        -------
        v_md : numpy.ndarray

        Notes
        -----
        Variance of microscopic mean diffusivities is calculated as

            .. math::

                V_\\text{MD} = \\mathbb{C} : \\mathbb{E}_\\text{bulk}
        """
        v_md = np.matmul(self.params[..., np.newaxis, 7::], from_6x6_to_21x1(E_bulk))[
            ..., 0, 0
        ]
        return v_md

    @auto_attr
    def v_shear(self):
        """Shear variance.

        Returns
        -------
        v_shear : numpy.ndarray

        Notes
        -----
        Shear variance is calculated as

            .. math::

                V_\\text{shear} = \\mathbb{C} : \\mathbb{E}_\\text{shear}
        """
        v_shear = np.matmul(
            self.params[..., np.newaxis, 7::], from_6x6_to_21x1(E_shear)
        )[..., 0, 0]
        return v_shear

    @auto_attr
    def v_iso(self):
        """Total isotropic variance.

        Returns
        -------
        v_iso : numpy.ndarray

        Notes
        -----
        Total isotropic variance is calculated as

            .. math::

                V_\\text{iso} = \\mathbb{C} : \\mathbb{E}_\\text{iso}
        """
        v_iso = np.matmul(self.params[..., np.newaxis, 7::], from_6x6_to_21x1(E_iso))[
            ..., 0, 0
        ]
        return v_iso

    @auto_attr
    def d_sq(self):
        """Diffusion tensor's outer product with itself.

        Returns
        -------
        d_sq : numpy.ndarray
        """
        d_sq = np.matmul(
            self.params[..., 1:7, np.newaxis], self.params[..., np.newaxis, 1:7]
        )
        return d_sq

    @auto_attr
    def mean_d_sq(self):
        """Average of microscopic diffusion tensors' outer products with
        themselves.

        Returns
        -------
        mean_d_sq : numpy.ndarray

        Notes
        -----
        Average of microscopic diffusion tensors' outer products with
        themselves is calculated as

            .. math::

                \\langle \\mathbf{D} \\otimes \\mathbf{D} \\rangle =
                \\mathbb{C} +
                \\langle \\mathbf{D} \\rangle \\otimes \\langle \\mathbf{D}
                \\rangle
        """
        mean_d_sq = from_21x1_to_6x6(self.params[..., 7::, np.newaxis]) + self.d_sq
        return mean_d_sq

    @auto_attr
    def c_md(self):
        """Normalized variance of mean diffusivities.

        Returns
        -------
        c_md : numpy.ndarray

        Notes
        -----
        Normalized variance of microscopic mean diffusivities is calculated as

            .. math::

                C_\\text{MD} = \\frac{\\mathbb{C} : \\mathbb{E}_\\text{bulk}}
                {\\langle \\mathbf{D} \\otimes \\mathbf{D} \\rangle :
                \\mathbb{E}_\\text{bulk}}
        """
        c_md = (
            self.v_md
            / np.matmul(
                np.swapaxes(from_6x6_to_21x1(self.mean_d_sq), -1, -2),
                from_6x6_to_21x1(E_bulk),
            )[..., 0, 0]
        )
        return c_md

    @auto_attr
    def c_mu(self):
        """Normalized microscopic anisotropy.

        Returns
        -------
        c_mu : numpy.ndarray

        Notes
        -----
        Normalized microscopic anisotropy is calculated as

            .. math::

                C_\\mu = \\frac{3}{2} \\frac{\\langle \\mathbf{D} \\otimes
                \\mathbf{D}
                \\rangle : \\mathbb{E}_\\text{shear}}{\\langle \\mathbf{D}
                \\otimes
                \\mathbf{D} \\rangle : \\mathbb{E}_\\text{iso}}
        """
        c_mu = (
            1.5
            * np.matmul(
                np.swapaxes(from_6x6_to_21x1(self.mean_d_sq), -1, -2),
                from_6x6_to_21x1(E_shear),
            )
            / np.matmul(
                np.swapaxes(from_6x6_to_21x1(self.mean_d_sq), -1, -2),
                from_6x6_to_21x1(E_iso),
            )
        )[..., 0, 0]
        return c_mu

    @auto_attr
    def ufa(self):
        """Microscopic fractional anisotropy.

        Returns
        -------
        ufa : numpy.ndarray

        Notes
        -----
        Microscopic fractional anisotropy is calculated as

            .. math::

                \\mu\\text{FA} = \\sqrt{C_\\mu}
        """
        ufa = np.sqrt(self.c_mu)
        return ufa

    @auto_attr
    def c_m(self):
        """Normalized macroscopic anisotropy.

        Returns
        -------
        c_m : numpy.ndarray

        Notes
        -----
        Normalized macroscopic anisotropy is calculated as

            .. math::

                C_\\text{M} = \\frac{3}{2} \\frac{\\langle \\mathbf{D} \\rangle
                \\otimes \\langle \\mathbf{D} \\rangle :
                \\mathbb{E}_\\text{shear}}
                {\\langle \\mathbf{D} \\rangle \\otimes \\langle \\mathbf{D}
                \\rangle :
                \\mathbb{E}_\\text{iso}}
        """
        c_m = (
            1.5
            * np.matmul(
                np.swapaxes(from_6x6_to_21x1(self.d_sq), -1, -2),
                from_6x6_to_21x1(E_shear),
            )
            / np.matmul(
                np.swapaxes(from_6x6_to_21x1(self.d_sq), -1, -2),
                from_6x6_to_21x1(E_iso),
            )
        )[..., 0, 0]
        return c_m

    @auto_attr
    def fa(self):
        """Fractional anisotropy.

        Returns
        -------
        fa : numpy.ndarray

        Notes
        -----
        Fractional anisotropy is calculated as

            .. math::

                \\text{FA} = \\sqrt{C_\\text{M}}
        """
        fa = np.sqrt(self.c_m)
        return fa

    @auto_attr
    def c_c(self):
        """Microscopic orientation coherence.

        Returns
        -------
        c_c : numpy.ndarray

        Notes
        -----
        Microscopic orientation coherence is calculated as

            .. math::

                C_c = \\frac{C_\\text{M}}{C_\\mu}
        """
        c_c = self.c_m / self.c_mu
        return c_c

    @auto_attr
    def mk(self):
        """Mean kurtosis.

        Returns
        -------
        mk : numpy.ndarray

        Notes
        -----
        Mean kurtosis is calculated as

            .. math::

                \\text{MK} = K_\\text{bulk} + K_\\text{shear}
        """
        mk = self.k_bulk + self.k_shear
        return mk

    @auto_attr
    def k_bulk(self):
        """Bulk kurtosis.

        Returns
        -------
        k_bulk : numpy.ndarray

        Notes
        -----
        Bulk kurtosis is calculated as

            .. math::

                K_\\text{bulk} = 3 \\frac{\\mathbb{C} :
                \\mathbb{E}_\\text{bulk}}
                {\\langle \\mathbf{D} \\rangle \\otimes \\langle \\mathbf{D}
                \\rangle : \\mathbb{E}_\\text{bulk}}
        """
        k_bulk = (
            3
            * np.matmul(self.params[..., np.newaxis, 7::], from_6x6_to_21x1(E_bulk))
            / np.matmul(
                np.swapaxes(from_6x6_to_21x1(self.d_sq), -1, -2),
                from_6x6_to_21x1(E_bulk),
            )
        )[..., 0, 0]
        return k_bulk

    @auto_attr
    def k_shear(self):
        """Shear kurtosis.

        Returns
        -------
        k_shear : numpy.ndarray

        Notes
        -----
        Shear kurtosis is calculated as

            .. math::

                K_\\text{shear} = \\frac{6}{5} \\frac{\\mathbb{C} :
                \\mathbb{E}_\\text{shear}}{\\langle \\mathbf{D} \\rangle
                \\otimes
                \\langle \\mathbf{D} \\rangle : \\mathbb{E}_\\text{bulk}}
        """
        k_shear = (
            6
            / 5
            * np.matmul(self.params[..., np.newaxis, 7::], from_6x6_to_21x1(E_shear))
            / np.matmul(
                np.swapaxes(from_6x6_to_21x1(self.d_sq), -1, -2),
                from_6x6_to_21x1(E_bulk),
            )
        )[..., 0, 0]
        return k_shear

    @auto_attr
    def k_mu(self):
        """Microscopic kurtosis.

        Returns
        -------
        k_mu : numpy.ndarray

        Notes
        -----
        Microscopic kurtosis is calculated as

            .. math::

                K_\\mu = \\frac{6}{5} \\frac{\\langle \\mathbf{D} \\otimes
                \\mathbf{D}
                \\rangle : \\mathbb{E}_\\text{shear}}{\\langle \\mathbf{D}
                \\rangle
                \\otimes \\langle \\mathbf{D} \\rangle :
                \\mathbb{E}_\\text{bulk}}
        """
        k_mu = (
            6
            / 5
            * np.matmul(
                np.swapaxes(from_6x6_to_21x1(self.mean_d_sq), -1, -2),
                from_6x6_to_21x1(E_shear),
            )
            / np.matmul(
                np.swapaxes(from_6x6_to_21x1(self.d_sq), -1, -2),
                from_6x6_to_21x1(E_bulk),
            )
        )[..., 0, 0]
        return k_mu


common_fit_methods = {"OLS": _ols_fit, "WLS": _wls_fit, "SDPdc": _sdpdc_fit}



"""Class for profiling cython code"""

import os
import subprocess

from dipy.testing.decorators import warning_for_keywords
from dipy.utils.optpkg import optional_package

cProfile, _, _ = optional_package("cProfile")
pstats, _, _ = optional_package(
    "pstats",
    trip_msg="pstats is not installed.  It is "
    "part of the python-profiler package in "
    "Debian/Ubuntu",
)


class Profiler:
    """Profile python/cython files or functions

    If you are profiling cython code you need to add
    # cython: profile=True on the top of your .pyx file

    and for the functions that you do not want to profile you can use
    this decorator in your cython files

    @cython.profile(False)

    Parameters
    ----------
    caller : file or function call
    args : function arguments

    Attributes
    ----------
    stats : function, stats.print_stats(10) will prin the 10 slower functions

    Examples
    --------
    from dipy.core.profile import Profiler
    import numpy as np
    p=Profiler(np.sum,np.random.rand(1000000,3))
    fname='test.py'
    p=Profiler(fname)
    p.print_stats(10)
    p.print_stats('det')

    References
    ----------
    https://docs.cython.org/src/tutorial/profiling_tutorial.html
    https://docs.python.org/library/profile.html
    https://github.com/rkern/line_profiler

    """

    def __init__(self, *args, call=None):
        # Delay import until use of class instance.  We were getting some very
        # odd build-as-we-go errors running tests and documentation otherwise
        import pyximport

        pyximport.install()

        try:
            ext = os.path.splitext(call)[1].lower()
            print("ext", ext)
            if ext in (".py", ".pyx"):  # python/cython file
                print("profiling python/cython file ...")
                subprocess.call(
                    ["python3", "-m", "cProfile", "-o", "profile.prof", call]
                )
                s = pstats.Stats("profile.prof")
                stats = s.strip_dirs().sort_stats("time")
                self.stats = stats

        except Exception:
            print("profiling function call ...")
            self.args = args
            self.call = call

            cProfile.runctx(
                "self._profile_function()", globals(), locals(), "profile.prof"
            )
            s = pstats.Stats("profile.prof")
            stats = s.strip_dirs().sort_stats("time")
            self.stats = stats

    def _profile_function(self):
        self.call(*self.args)

    @warning_for_keywords()
    def print_stats(self, *, N=10):
        """Print stats for profiling

        You can use it in all different ways developed in pstats
        for example
        print_stats(10) will give you the 10 slowest calls
        or
        print_stats('function_name')
        will give you the stats for all the calls with name 'function_name'

        Parameters
        ----------
        N : stats.print_stats argument

        """
        self.stats.print_stats(N)



"""Testing volumes"""

import numpy as np
import numpy.testing as npt

from dipy.utils.volume import adjacency_calc


def test_adjacency_calc():
    """
    Test adjacency_calc function, which calculates indices of adjacent voxels
    """

    cutoff = 1.99
    for img_shape in [(50, 50), (50, 50, 5)]:
        mask = None
        adj = adjacency_calc(img_shape, mask=mask, cutoff=cutoff)
        # check that adj in the first voxel is correct
        adj[0].sort()
        if len(img_shape) == 2:
            npt.assert_equal(adj[0], [0, 1, 50, 51])
        if len(img_shape) == 3:
            npt.assert_equal(adj[0], [0, 1, 5, 6, 250, 251, 255, 256])

        mask = np.zeros(img_shape, dtype=int)
        if len(img_shape) == 2:
            mask[10:40, 20:30] = 1
        if len(img_shape) == 3:
            mask[10:40, 20:30, :] = 1
        adj = adjacency_calc(img_shape, mask=mask, cutoff=cutoff)
        # check that adj in the first voxel is correct
        if len(img_shape) == 2:
            npt.assert_equal(adj[0], [0, 1, 10, 11])
        if len(img_shape) == 3:
            npt.assert_equal(adj[0], [0, 1, 5, 6, 50, 51, 55, 56])
        # test that adj only corresponds to flattened and masked data
        npt.assert_equal(len(adj), mask.sum())



import warnings

import numpy as np
import numpy.testing as npt

from dipy.core.sphere import Sphere
from dipy.denoise.enhancement_kernel import EnhancementKernel
from dipy.denoise.shift_twist_convolution import convolve, convolve_sf
from dipy.reconst.shm import descoteaux07_legacy_msg, sf_to_sh, sh_to_sf


def test_enhancement_kernel():
    """Test if the kernel values are correct by comparison against the values
    originally calculated by implementation in Mathematica, and at the same time
    checks the symmetry of the kernel."""

    D33 = 1.0
    D44 = 0.04
    t = 1
    k = EnhancementKernel(D33, D44, t, orientations=0, force_recompute=True)

    y = np.array([0.0, 0.0, 0.0])
    v = np.array([0.0, 0.0, 1.0])
    orientationlist = [
        [0.0, 0.0, 1.0],
        [-0.0527864, 0.688191, 0.723607],
        [-0.67082, -0.16246, 0.723607],
        [-0.0527864, -0.688191, 0.723607],
        [0.638197, -0.262866, 0.723607],
        [0.831052, 0.238856, 0.502295],
        [0.262866, -0.809017, -0.525731],
        [0.812731, 0.295242, -0.502295],
        [-0.029644, 0.864188, -0.502295],
        [-0.831052, 0.238856, -0.502295],
        [-0.638197, -0.262866, -0.723607],
        [-0.436009, 0.864188, -0.251148],
        [-0.687157, -0.681718, 0.251148],
        [0.67082, -0.688191, 0.276393],
        [0.67082, 0.688191, 0.276393],
        [0.947214, 0.16246, -0.276393],
        [-0.861803, -0.425325, -0.276393],
    ]
    positionlist = [
        [-0.108096, 0.0412229, 0.339119],
        [0.220647, -0.422053, 0.427524],
        [-0.337432, -0.0644619, -0.340777],
        [0.172579, -0.217602, -0.292446],
        [-0.271575, -0.125249, -0.350906],
        [-0.483807, 0.326651, 0.191993],
        [-0.480936, -0.0718426, 0.33202],
        [0.497193, -0.00585659, -0.251344],
        [0.237737, 0.013634, -0.471988],
        [0.367569, -0.163581, 0.0723955],
        [0.47859, -0.143252, 0.318579],
        [-0.21474, -0.264929, -0.46786],
        [-0.0684234, 0.0342464, 0.0942475],
        [0.344272, 0.423119, -0.303866],
        [0.0430714, 0.216233, -0.308475],
        [0.386085, 0.127333, 0.0503609],
        [0.334723, 0.071415, 0.403906],
    ]
    kernelvalues = [
        0.10701063104295713,
        0.0030052117308328923,
        0.003125410084676201,
        0.0031765819772012613,
        0.003127254657020615,
        0.0001295130396491743,
        6.882352014430076e-14,
        1.3821277371353332e-13,
        1.3951939946082493e-13,
        1.381612071786285e-13,
        5.0861109163441125e-17,
        1.0722120295517027e-10,
        2.425145934791457e-6,
        3.557919265806602e-6,
        3.6669510385105265e-6,
        5.97473789679846e-11,
        6.155412262223178e-11,
    ]

    for p in range(len(orientationlist)):
        r = np.array(orientationlist[p])
        x = np.array(positionlist[p])
        npt.assert_almost_equal(k.evaluate_kernel(x, y, r, v), kernelvalues[p])


def test_spike():
    """Test if a convolution with a delta spike is equal to the kernel
    saved in the lookup table."""

    # create kernel
    D33 = 1.0
    D44 = 0.04
    t = 1
    num_orientations = 5
    k = EnhancementKernel(
        D33, D44, t, orientations=num_orientations, force_recompute=True
    )

    # create a delta spike
    numorientations = k.get_orientations().shape[0]
    spike = np.zeros((7, 7, 7, numorientations), dtype=np.float64)
    spike[3, 3, 3, 0] = 1

    # convolve kernel with delta spike
    csd_enh = convolve_sf(spike, k, test_mode=True, normalize=False)

    # check if kernel matches with the convolved delta spike
    totalsum = 0.0
    for i in range(0, numorientations):
        totalsum += np.sum(
            np.array(k.get_lookup_table())[i, 0, :, :, :]
            - np.array(csd_enh)[:, :, :, i]
        )
    npt.assert_equal(totalsum, 0.0)


def test_normalization():
    """Test the normalization routine applied after a convolution"""
    # create kernel
    D33 = 1.0
    D44 = 0.04
    t = 1
    num_orientations = 5
    k = EnhancementKernel(
        D33, D44, t, orientations=num_orientations, force_recompute=True
    )

    # create a constant dataset
    numorientations = k.get_orientations().shape[0]
    spike = np.ones((7, 7, 7, numorientations), dtype=np.float64)

    # convert dataset to SH
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        spike_sh = sf_to_sh(spike, k.get_sphere(), sh_order_max=8)

        # convolve kernel with delta spike and apply normalization
        csd_enh = convolve(spike_sh, k, sh_order_max=8, test_mode=True, normalize=True)

        # convert dataset to DSF
        csd_enh_dsf = sh_to_sf(csd_enh, k.get_sphere(), sh_order_max=8, basis_type=None)

    # test if the normalization is performed correctly
    npt.assert_almost_equal(np.amax(csd_enh_dsf), np.amax(spike))


def test_kernel_input():
    """Test the kernel for inputs of type Sphere, type int and for input None"""

    sph = Sphere(x=1, y=0, z=0)
    D33 = 1.0
    D44 = 0.04
    t = 1
    k = EnhancementKernel(D33, D44, t, orientations=sph, force_recompute=True)
    npt.assert_equal(k.get_lookup_table().shape, (1, 1, 7, 7, 7))

    num_orientations = 2
    k = EnhancementKernel(
        D33, D44, t, orientations=num_orientations, force_recompute=True
    )
    npt.assert_equal(k.get_lookup_table().shape, (2, 2, 7, 7, 7))

    k = EnhancementKernel(D33, D44, t, orientations=0, force_recompute=True)
    npt.assert_equal(k.get_lookup_table().shape, (0, 0, 7, 7, 7))



# A type of -*- python -*- file
"""This module contains the parts of dipy.tracking.utils that need to be
implemented in cython.
"""
import cython

cdef extern from "dpy_math.h" nogil:
    double fmin(double x, double y)
from libc.math cimport ceil, floor, fabs, sqrt

import numpy as np
cimport numpy as cnp
from dipy.tracking._utils import _mapping_to_voxel, _to_voxel_coordinates


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.profile(False)
def _voxel2streamline(sl,
                      cnp.ndarray[cnp.npy_intp, ndim=2] unique_idx):
    """
    Maps voxels to streamlines and streamlines to voxels, for setting up
    the LiFE equations matrix

    Parameters
    ----------
    sl : list
        A collection of streamlines, each n by 3, with n being the number of
        nodes in the fiber.

    unique_idx : array.
       The unique indices in the streamlines

    Returns
    -------
    v2f, v2fn : tuple of dicts

    The first dict in the tuple answers the question: Given a voxel (from
    the unique indices in this model), which fibers pass through it?

    The second answers the question: Given a streamline, for each voxel that
    this streamline passes through, which nodes of that streamline are in that
    voxel?
    """
    # Define local counters:
    cdef int s_idx, node_idx, voxel_id, ii
    cdef dict vox_dict = {}
    for ii in range(len(unique_idx)):
        vox = unique_idx[ii]
        vox_dict[vox[0], vox[1], vox[2]] = ii
    # Outputs are these dicts:
    cdef dict v2f = {}
    cdef dict v2fn = {}
    # In each fiber:
    for s_idx in range(len(sl)):
        sl_as_idx = np.round(sl[s_idx]).astype(int)
        v2fn[s_idx] = {}
        # In each voxel present in there:
        for node_idx in range(len(sl_as_idx)):
            node = sl_as_idx[node_idx]
            # What serial number is this voxel in the unique voxel indices:
            voxel_id = vox_dict[node[0], node[1], node[2]]
            # Add that combination to the dict:
            if voxel_id in v2f:
                if s_idx not in v2f[voxel_id]:
                    v2f[voxel_id].append(s_idx)
            else:
                v2f[voxel_id] = [s_idx]
            # All the nodes going through this voxel get its number:
            if voxel_id in v2fn[s_idx]:
                v2fn[s_idx][voxel_id].append(node_idx)
            else:
                v2fn[s_idx][voxel_id] = [node_idx]
    return v2f ,v2fn


def streamline_mapping(streamlines, affine=None,
                       mapping_as_streamlines=False):
    """Creates a mapping from voxel indices to streamlines.

    Returns a dictionary where each key is a 3d voxel index and the associated
    value is a list of the streamlines that pass through that voxel.

    Parameters
    ----------
    streamlines : sequence
        A sequence of streamlines.
    affine : array_like (4, 4), optional
        The mapping from voxel coordinates to streamline coordinates.
        The streamline values are assumed to be in voxel coordinates.
        IE ``[0, 0, 0]`` is the center of the first voxel and the voxel size
        is ``[1, 1, 1]``.
    mapping_as_streamlines : bool, optional, False by default
        If True voxel indices map to lists of streamline objects. Otherwise
        voxel indices map to lists of integers.

    Returns
    -------
    mapping : defaultdict(list)
        A mapping from voxel indices to the streamlines that pass through that
        voxel.

    Examples
    --------
    >>> streamlines = [np.array([[0., 0., 0.],
    ...                          [1., 1., 1.],
    ...                          [2., 3., 4.]]),
    ...                np.array([[0., 0., 0.],
    ...                          [1., 2., 3.]])]
    >>> mapping = streamline_mapping(streamlines, affine=np.eye(4))
    >>> mapping[0, 0, 0]
    [0, 1]
    >>> mapping[1, 1, 1]
    [0]
    >>> mapping[1, 2, 3]
    [1]
    >>> mapping.get((3, 2, 1), 'no streamlines')
    'no streamlines'
    >>> mapping = streamline_mapping(streamlines, affine=np.eye(4),
    ...                              mapping_as_streamlines=True)
    >>> mapping[1, 2, 3][0] is streamlines[1]
    True

    """
    cdef:
        cnp.npy_intp[:, :] voxel_indices

    lin, offset = _mapping_to_voxel(affine)
    if mapping_as_streamlines:
        streamlines = list(streamlines)
    mapping = {}

    for i, sl in enumerate(streamlines):
        voxel_indices = _to_voxel_coordinates(sl, lin, offset)

        # Get the unique voxels every streamline passes through
        uniq_points = set()
        for j in range(voxel_indices.shape[0]):
            point = (voxel_indices[j, 0],
                     voxel_indices[j, 1],
                     voxel_indices[j, 2])
            uniq_points.add(point)

        # Add the index of this streamline for each unique voxel
        for point in uniq_points:
            if point in mapping:
                mapping[point].append(i)
            else:
                mapping[point] = [i]

    # If mapping_as_streamlines replace ids with streamlines
    if mapping_as_streamlines:
        for key in mapping:
            mapping[key] = [streamlines[i] for i in mapping[key]]

    return mapping


@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline cnp.double_t norm(cnp.double_t x,
                              cnp.double_t y,
                              cnp.double_t z) noexcept nogil:
    cdef cnp.double_t val = sqrt(x*x + y*y + z*z)
    return val


# Changing this to a memview was slower.
@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline void c_get_closest_edge(cnp.double_t* p,
                                    cnp.double_t* direction,
                                    cnp.double_t* edge,
                                    double eps=1.) noexcept nogil:
     edge[0] = floor(p[0] + eps) if direction[0] >= 0.0 else ceil(p[0] - eps)
     edge[1] = floor(p[1] + eps) if direction[1] >= 0.0 else ceil(p[1] - eps)
     edge[2] = floor(p[2] + eps) if direction[2] >= 0.0 else ceil(p[2] - eps)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def _streamlines_in_mask(list streamlines,
                         cnp.uint8_t[:,:,:] mask,
                         lin_T, offset):
    """Filters streamlines based on whether or not they pass through a ROI,
    using a line-based algorithm for compressed streamlines.

    This function is private because it's supposed to be called only by
    tracking.utils.target_line_based.

    Parameters
    ----------
    streamlines : sequence
        A sequence of streamlines.
    mask : array-like (uint8)
        A mask used as a target. Non-zero values are considered to be within
        the target region.
    lin_T : array (3, 3)
        Transpose of the linear part of the mapping to voxel space. Obtained
        with `_mapping_to_voxel`.
    offset : array or scalar
        Mapping to voxel space. Obtained with `_mapping_to_voxel`.

    Returns
    -------
    in_mask : 1D array of bool (uint8), one for each streamline.
        0 if passing through mask, 1 otherwise
        (2 for single-point streamline)
    """
    cdef cnp.double_t[:,:] voxel_indices

    cdef cnp.npy_intp nb_streamlines = len(streamlines)
    cdef cnp.uint8_t[:] in_mask = np.zeros(nb_streamlines, dtype=np.uint8)
    cdef cnp.npy_intp streamline_idx

    for streamline_idx in range(nb_streamlines):
        # Can't call _to_voxel_coordinates because it casts to int
        voxel_indices = np.dot(streamlines[streamline_idx], lin_T) + offset

        in_mask[streamline_idx] = _streamline_in_mask(voxel_indices, mask)

    return np.asarray(in_mask)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef cnp.npy_intp _streamline_in_mask(
        cnp.double_t[:,:] streamline,
        cnp.uint8_t[:,:,:] mask) noexcept nogil:
    """
    Check if a single streamline is passing through a mask. This is an utility
    function to make streamlines_in_mask() more readable.
    """
    cdef cnp.double_t *current_pt = [0.0, 0.0, 0.0]
    cdef cnp.double_t *next_pt = [0.0, 0.0, 0.0]
    cdef cnp.double_t *direction = [0.0, 0.0, 0.0]
    cdef cnp.double_t *current_edge = [0.0, 0.0, 0.0]

    cdef cnp.double_t direction_norm, remaining_distance
    cdef cnp.double_t length_ratio, half_ratio
    cdef cnp.npy_intp point_idx, dim_idx
    cdef cnp.npy_intp x, y, z

    if streamline.shape[0] <= 1:
        return 2  # Single-point streamline

    # This loop is time-critical
    # Changed to -1 because we get the next point in the loop
    for point_idx in range(streamline.shape[0] - 1):
        # Assign current and next point, find vector between both,
        # and use the current point as nearest edge for testing.
        for dim_idx in range(3):
            current_pt[dim_idx] = streamline[point_idx, dim_idx]
            next_pt[dim_idx] = streamline[point_idx + 1, dim_idx]
            direction[dim_idx] = next_pt[dim_idx] - current_pt[dim_idx]
            current_edge[dim_idx] = current_pt[dim_idx]

        # Set the "remaining_distance" var to compute remaining length of
        # vector to process
        direction_norm = norm(direction[0], direction[1], direction[2])
        remaining_distance = direction_norm

        # If consecutive coordinates are the same, skip one.
        if direction_norm == 0:
            continue

        # Check if it's already a real edge. If not, find the closest edge.
        if floor(current_edge[0]) != current_edge[0] and \
           floor(current_edge[1]) != current_edge[1] and \
           floor(current_edge[2]) != current_edge[2]:
            # All coordinates are not "integers", and therefore, not on the
            # edge. Fetch the closest edge.
            c_get_closest_edge(current_pt, direction, current_edge)

        while True:
            # Compute the smallest ratio of direction's length to get to an
            # edge. This effectively means we find the first edge
            # encountered
            # Set large value for length_ratio
            length_ratio = 10000
            for dim_idx in range(3):
                if direction[dim_idx] != 0:
                    length_ratio = fmin(
                        fabs((current_edge[dim_idx] - current_pt[dim_idx])
                            / direction[dim_idx]),
                        length_ratio)

            # Check if last point is already on an edge
            remaining_distance -= length_ratio * direction_norm
            if remaining_distance < 0 and not fabs(remaining_distance) < 1e-8:
                break

            # Find the coordinates of voxel containing current point, to
            # tag it in the map
            half_ratio = 0.5 * length_ratio
            x = <cnp.npy_intp>floor(current_pt[0] + half_ratio * direction[0])
            y = <cnp.npy_intp>floor(current_pt[1] + half_ratio * direction[1])
            z = <cnp.npy_intp>floor(current_pt[2] + half_ratio * direction[2])
            if 0 <= x < mask.shape[0] and 0 <= y < mask.shape[1] and 0 <= z < mask.shape[2]:
                if mask[x, y, z]:
                    return 1

            # current_pt is moved to the closest edge
            for dim_idx in range(3):
                current_pt[dim_idx] = \
                    length_ratio * direction[dim_idx] + current_pt[dim_idx]

                # Snap really small values to 0.
                if fabs(current_pt[dim_idx]) <= 1e-16:
                    current_pt[dim_idx] = 0.0

            c_get_closest_edge(current_pt, direction, current_edge)

    # Check last point
    x = <cnp.npy_intp>floor(next_pt[0])
    y = <cnp.npy_intp>floor(next_pt[1])
    z = <cnp.npy_intp>floor(next_pt[2])
    if 0 <= x < mask.shape[0] and 0 <= y < mask.shape[1] and 0 <= z < mask.shape[2]:
        if mask[x, y, z]:
            return 1
    return 0


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.profile(False)
def track_counts(tracks, vol_dims, vox_sizes=(1,1,1), return_elements=True):
    """ Counts of points in `tracks` that pass through voxels in volume

    We find whether a point passed through a track by rounding the mm
    point values to voxels.  For a track that passes through a voxel more
    than once, we only record counts and elements for the first point in
    the line that enters the voxel.

    Parameters
    ----------
    tracks : sequence
       sequence of T tracks.  One track is an ndarray of shape (N, 3), where N
       is the number of points in that track, and ``tracks[t][n]`` is the n-th
       point in the t-th track.  Points are of form x, y, z in *voxel mm*
       coordinates. That is, if ``i, j, k`` is the possibly non-integer voxel
       coordinate of the track point, and `vox_sizes` are 3 floats giving voxel
       sizes of dimensions 0, 1, 2 respectively, then the voxel mm coordinates
       ``x, y, z`` are simply ``i * vox_sizes[0], j * vox_sizes[1], k *
       vox_sizes[2]``.  This convention derives from trackviz.  To pass in
       tracks as voxel coordinates, just pass ``vox_sizes=(1, 1, 1)`` (see
       below).
    vol_dims : sequence length 3
       volume dimensions in voxels, x, y, z.
    vox_sizes : optional, sequence length 3
       voxel sizes in mm.  Default is (1,1,1)
    return_elements : {True, False}, optional
       If True, also return object array with one list per voxel giving
       track indices and point indices passing through the voxel (see
       below)

    Returns
    -------
    tcs : ndarray shape `vol_dim`
       An array where entry ``tcs[x, y, z]`` is the number of tracks
       that passed through voxel at voxel coordinate x, y, z
    tes : ndarray dtype np.object, shape `vol_dim`
       If `return_elements` is True, we also return an object array with
       one object per voxel. The objects at each voxel are a list of
       integers, where the integers are the indices of the track that
       passed through the voxel.

    Examples
    --------
    Imagine you have a volume (voxel) space of dimension ``(10,20,30)``.
    Imagine you had voxel coordinate tracks in ``vs``.  To just fill an array
    with the counts of how many tracks pass through each voxel:

    >>> vox_track0 = np.array([[0, 0, 0], [1.1, 2.2, 3.3], [2.2, 4.4, 6.6]])
    >>> vox_track1 = np.array([[0, 0, 0], [0, 0, 1], [0, 0, 2]])
    >>> vs = (vox_track0, vox_track1)
    >>> vox_dim = (10, 20, 30) # original voxel array size
    >>> tcs = track_counts(vs, vox_dim, (1, 1, 1), False)
    >>> tcs.shape
    (10, 20, 30)
    >>> tcs[0, 0, 0:4]
    array([2, 1, 1, 0])
    >>> tcs[1, 2, 3], tcs[2, 4, 7]
    (1, 1)

    You can also use the routine to count into larger-than-voxel boxes.  To do
    this, increase the voxel size and decrease the ``vox_dim`` accordingly:

    >>> tcs=track_counts(vs, (10/2., 20/2., 30/2.), (2,2,2), False)
    >>> tcs.shape
    (5, 10, 15)
    >>> tcs[1,1,2], tcs[1,2,3]
    (1, 1)
    """
    vol_dims = np.asarray(vol_dims).astype(int)
    vox_sizes = np.asarray(vox_sizes).astype(np.double)
    n_voxels = np.prod(vol_dims)
    # output track counts array, flattened
    cdef cnp.ndarray[cnp.npy_intp, ndim=1] tcs = \
        np.zeros((n_voxels,), dtype=np.intp)
    # pointer to output track indices
    cdef cnp.npy_intp i
    if return_elements:
        el_inds = {}
    # cython numpy pointer to individual track array
    cdef cnp.ndarray[cnp.float_t, ndim=2] t
    # cython numpy pointer to point in track array
    cdef cnp.ndarray[cnp.float_t, ndim=1] in_pt
    # processed point
    cdef int out_pt[3]
    # various temporary loop and working variables
    cdef int tno, pno, cno
    cdef cnp.npy_intp el_no, v
    # fill native C arrays from inputs
    cdef int vd[3]
    cdef double vxs[3]
    for cno in range(3):
        vd[cno] = vol_dims[cno]
        vxs[cno] = vox_sizes[cno]
    # return_elements to C native
    cdef int ret_elf = <int>return_elements
    # x slice size (C array ordering)
    cdef cnp.npy_intp yz = vd[1] * vd[2]
    for tno in range(len(tracks)):
        t = tracks[tno].astype(float)
        # set to find unique voxel points in track
        in_inds = set()
        # the loop below is time-critical
        for pno in range(cnp.PyArray_DIM(t, 0)):
            in_pt = t[pno]
            # Round to voxel coordinates, and set coordinates outside
            # volume to volume edges
            for cno in range(3):
                v = <int>floor(in_pt[cno] / vxs[cno] + 0.5)
                if v < 0:
                    v = 0
                elif v >= vd[cno]:
                    v = vd[cno]-1 # last index for this dimension
                out_pt[cno] = v
            # calculate element number in flattened tcs array
            el_no = out_pt[0] * yz + out_pt[1] * vd[2] + out_pt[2]
            # discard duplicates
            if el_no in in_inds:
                continue
            in_inds.add(el_no)
            # set elements into object array
            if ret_elf:
                key = (out_pt[0], out_pt[1], out_pt[2])
                val = tno
                if tcs[el_no]:
                    el_inds[key].append(val)
                else:
                    el_inds[key] = [val]
            # set value into counts
            tcs[el_no] += 1
    if ret_elf:
        return tcs.reshape(vol_dims), el_inds
    return tcs.reshape(vol_dims)



"""
==========================================================
Enhancing QuickBundles with different metrics and features
==========================================================

QuickBundles :footcite:p:`Garyfallidis2012a` is a flexible algorithm that
requires only a distance metric and an adjacency threshold to perform
clustering. There is a wide variety of metrics that could be used to cluster
streamlines.

The purpose of this tutorial is to show how to easily create new ``Feature``
and new ``Metric`` classes that can be used by QuickBundles.

.. _clustering-framework:

Clustering framework
====================
DIPY_ provides a simple, flexible and fast framework to do clustering of
sequential data (e.g. streamlines).

A *sequential datum* in DIPY is represented as a numpy array of size
:math:`(N \times D)`, where each row of the array represents a $D$ dimensional
point of the sequence. A set of these sequences is represented as a list of
numpy arrays of size :math:`(N_i \times D)` for :math:`i=1:M` where $M$ is the
number of sequences in the set.

This clustering framework is modular and divided in three parts:

#. Feature extraction

#. Distance computation

#. Clustering algorithm

The **feature extraction** part includes any preprocessing needed to be done on
the data before computing distances between them (e.g. resampling the number of
points of a streamline). To define a new way of extracting features, one has to
subclass ``Feature`` (see below).

The **distance computation** part includes any metric capable of evaluating a
distance between two sets of features previously extracted from the data. To
define a new way of extracting features, one has to subclass ``Metric`` (see
below).

The **clustering algorithm** part represents the clustering algorithm itself
(e.g. QuickBundles, K-means, Hierarchical Clustering). More precisely, it
includes any algorithms taking as input a list of sequential data and
outputting a ``ClusterMap`` object.


Extending `Feature`
===================
This section will guide you through the creation of a new feature extraction
method that can be used in the context of this clustering framework. For a
list of available features in DIPY see
:ref:`sphx_glr_examples_built_segmentation_segment_clustering_features.py`.

Assuming a set of streamlines, the type of features we want to extract is the
arc length (i.e. the sum of the length of each segment for a given streamline).

Let's start by importing the necessary modules.
"""

import numpy as np

from dipy.data import get_fnames
from dipy.io.streamline import load_tractogram
from dipy.segment.clustering import QuickBundles
from dipy.segment.featurespeed import Feature, VectorOfEndpointsFeature
from dipy.segment.metric import Metric, SumPointwiseEuclideanMetric
from dipy.tracking.streamline import Streamlines, length
from dipy.viz import actor, colormap, window

###############################################################################
# We now define the class ``ArcLengthFeature`` that will perform the desired
# feature extraction. When subclassing ``Feature``, two methods have to be
# redefined: ``infer_shape`` and ``extract``.
#
# Also, an important property about feature extraction is whether or not
# its process is invariant to the order of the points within a streamline.
# This is needed as there is no way one can tell which extremity of a
# streamline is the beginning and which one is the end.


class ArcLengthFeature(Feature):
    """Computes the arc length of a streamline."""

    def __init__(self):
        # The arc length stays the same even if the streamline is reversed.
        super(ArcLengthFeature, self).__init__(is_order_invariant=True)

    def infer_shape(self, streamline):
        """Infers the shape of features extracted from `streamline`."""
        # Arc length is a scalar
        return 1

    def extract(self, streamline):
        """Extracts features from `streamline`."""
        return length(streamline)


###############################################################################
# The new feature extraction ``ArcLengthFeature`` is ready to be used. Let's
# use it to cluster a set of streamlines by their arc length. For educational
# purposes we will try to cluster a small streamline bundle known from
# neuroanatomy as the fornix.
#
# We start by loading the fornix streamlines.

fname = get_fnames(name="fornix")
fornix = load_tractogram(fname, "same", bbox_valid_check=False).streamlines

streamlines = Streamlines(fornix)

###############################################################################
# Perform QuickBundles clustering using the metric
# ``SumPointwiseEuclideanMetric`` and our ``ArcLengthFeature``.

metric = SumPointwiseEuclideanMetric(feature=ArcLengthFeature())
qb = QuickBundles(threshold=2.0, metric=metric)
clusters = qb.cluster(streamlines)

###############################################################################
# We will now visualize the clustering result.

# Color each streamline according to the cluster they belong to.
cmap = colormap.create_colormap(np.ravel(clusters.centroids))
colormap_full = np.ones((len(streamlines), 3))
for cluster, color in zip(clusters, cmap):
    colormap_full[cluster.indices] = color

scene = window.Scene()
scene.SetBackground(1, 1, 1)
scene.add(actor.streamtube(streamlines, colors=colormap_full))
window.record(scene=scene, out_path="fornix_clusters_arclength.png", size=(600, 600))

# Enables/disables interactive visualization
interactive = False
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Showing the different clusters obtained by using the arc length.
#
#
# Extending `Metric`
# ==================
# This section will guide you through the creation of a new metric that can be
# used in the context of this clustering framework. For a list of available
# metrics in DIPY see
# :ref:`sphx_glr_examples_built_segmentation_segment_clustering_metrics.py`.
#
# Assuming a set of streamlines, we want a metric that computes the cosine
# distance giving the vector between endpoints of each streamline (i.e. one
# minus the cosine of the angle between two vectors). For more information
# about this distance check
# `<https://en.wikipedia.org/wiki/Cosine_similarity>`_.
#
# We now define the class ``CosineMetric`` that will perform the desired
# distance computation. When subclassing ``Metric``, two methods have to be
# redefined: ``are_compatible`` and ``dist``. Moreover, when implementing the
# ``dist`` method, one needs to make sure the distance returned is symmetric
# (i.e. `dist(A, B) == dist(B, A)`).


class CosineMetric(Metric):
    """Compute the cosine distance between two streamlines."""

    def __init__(self):
        # For simplicity, features will be the vector between endpoints of a
        # streamline.
        super(CosineMetric, self).__init__(feature=VectorOfEndpointsFeature())

    def are_compatible(self, shape1, shape2):
        """Check if two features are vectors of same dimension.

        Basically this method exists so that we don't have to check
        inside the `dist` method (speedup).
        """
        return shape1 == shape2 and shape1[0] == 1

    def dist(self, v1, v2):
        """Compute a the cosine distance between two vectors."""

        def norm(x):
            return np.sqrt(np.sum(x**2))

        cos_theta = np.dot(v1, v2.T) / (norm(v1) * norm(v2))

        # Make sure it's in [-1, 1], i.e. within domain of arccosine
        cos_theta = np.minimum(cos_theta, 1.0)
        cos_theta = np.maximum(cos_theta, -1.0)
        return np.arccos(cos_theta) / np.pi  # Normalized cosine distance


###############################################################################
# The new distance ``CosineMetric`` is ready to be used. Let's use
# it to cluster a set of streamlines according to the cosine distance of the
# vector between their endpoints. For educational purposes we will try to
# cluster a small streamline bundle known from neuroanatomy as the fornix.
#
# We start by loading the fornix streamlines.

fname = get_fnames(name="fornix")
fornix = load_tractogram(fname, "same", bbox_valid_check=False)
streamlines = fornix.streamlines

###############################################################################
# Perform QuickBundles clustering using our metric ``CosineMetric``.

metric = CosineMetric()
qb = QuickBundles(threshold=0.1, metric=metric)
clusters = qb.cluster(streamlines)

###############################################################################
# We will now visualize the clustering result.

# Color each streamline according to the cluster they belong to.
cmap = colormap.create_colormap(np.arange(len(clusters)))
colormap_full = np.ones((len(streamlines), 3))
for cluster, color in zip(clusters, cmap):
    colormap_full[cluster.indices] = color

scene = window.Scene()
scene.SetBackground(1, 1, 1)
scene.add(actor.streamtube(streamlines, colors=colormap_full))
window.record(scene=scene, out_path="fornix_clusters_cosine.png", size=(600, 600))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Showing the different clusters obtained by using the cosine metric.
#
#
#
# References
# ----------
#
# .. footbibliography::
#



from os.path import join
from tempfile import TemporaryDirectory
import warnings

import numpy as np
from numpy.testing import assert_allclose, assert_equal

from dipy.data import get_fnames
from dipy.io.image import load_nifti, load_nifti_data, save_nifti
from dipy.io.peaks import load_pam
from dipy.reconst.shm import descoteaux07_legacy_msg
from dipy.workflows.reconst import ReconstDtiFlow


def test_reconst_dti_wls():
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        reconst_flow_core(ReconstDtiFlow)


def test_reconst_dti_nlls():
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        reconst_flow_core(ReconstDtiFlow, extra_args=[], extra_kwargs={})


def test_reconst_dti_alt_tensor():
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=descoteaux07_legacy_msg,
            category=PendingDeprecationWarning,
        )
        reconst_flow_core(
            ReconstDtiFlow, extra_args=[], extra_kwargs={"nifti_tensor": False}
        )


def reconst_flow_core(flow, extra_args=None, extra_kwargs=None):
    extra_args = extra_args or []
    extra_kwargs = extra_kwargs or {}

    with TemporaryDirectory() as out_dir:
        data_path, bval_path, bvec_path = get_fnames(name="small_25")
        volume, affine = load_nifti(data_path)
        mask = np.ones_like(volume[:, :, :, 0], dtype=np.uint8)
        mask_path = join(out_dir, "tmp_mask.nii.gz")
        save_nifti(mask_path, mask, affine)

        dti_flow = flow()

        args = [data_path, bval_path, bvec_path, mask_path]
        args.extend(extra_args)
        kwargs = {"out_dir": out_dir, "extract_pam_values": True}
        kwargs.update(extra_kwargs)

        dti_flow.run(*args, **kwargs)

        fa_path = dti_flow.last_generated_outputs["out_fa"]
        fa_data = load_nifti_data(fa_path)
        assert_equal(fa_data.shape, volume.shape[:-1])

        tensor_path = dti_flow.last_generated_outputs["out_tensor"]
        tensor_data = load_nifti_data(tensor_path)
        # Per default, tensor data is 5D, with six tensor elements on the last
        # dimension, except if nifti_tensor is set to False:
        if extra_kwargs.get("nifti_tensor", True):
            assert_equal(tensor_data.shape[-1], 6)
            assert_equal(tensor_data.shape[:-2], volume.shape[:-1])
        else:
            assert_equal(tensor_data.shape[-1], 6)
            assert_equal(tensor_data.shape[:-1], volume.shape[:-1])

        for out_name in ["out_ga", "out_md", "out_ad", "out_rd", "out_mode"]:
            out_path = dti_flow.last_generated_outputs[out_name]
            out_data = load_nifti_data(out_path)
            assert_equal(out_data.shape, volume.shape[:-1])

        rgb_path = dti_flow.last_generated_outputs["out_rgb"]
        rgb_data = load_nifti_data(rgb_path)
        assert_equal(rgb_data.shape[-1], 3)
        assert_equal(rgb_data.shape[:-1], volume.shape[:-1])

        evecs_path = dti_flow.last_generated_outputs["out_evec"]
        evecs_data = load_nifti_data(evecs_path)
        assert_equal(evecs_data.shape[-2:], (3, 3))
        assert_equal(evecs_data.shape[:-2], volume.shape[:-1])

        evals_path = dti_flow.last_generated_outputs["out_eval"]
        evals_data = load_nifti_data(evals_path)
        assert_equal(evals_data.shape[-1], 3)
        assert_equal(evals_data.shape[:-1], volume.shape[:-1])

        peaks_dir_path = dti_flow.last_generated_outputs["out_peaks_dir"]
        peaks_dir_data = load_nifti_data(peaks_dir_path)
        assert_equal(peaks_dir_data.shape[-1], 3)
        assert_equal(peaks_dir_data.shape[:-1], volume.shape[:-1])

        peaks_idx_path = dti_flow.last_generated_outputs["out_peaks_indices"]
        peaks_idx_data = load_nifti_data(peaks_idx_path)
        assert_equal(peaks_idx_data.shape[-1], 1)
        assert_equal(peaks_idx_data.shape[:-1], volume.shape[:-1])

        peaks_vals_path = dti_flow.last_generated_outputs["out_peaks_values"]
        peaks_vals_data = load_nifti_data(peaks_vals_path)
        assert_equal(peaks_vals_data.shape[-1], 1)
        assert_equal(peaks_vals_data.shape[:-1], volume.shape[:-1])

        pam = load_pam(dti_flow.last_generated_outputs["out_pam"])
        assert_allclose(pam.peak_dirs.reshape(peaks_dir_data.shape), peaks_dir_data)
        assert_allclose(pam.peak_values, peaks_vals_data)
        assert_allclose(pam.peak_indices, peaks_idx_data)



import re


def dollars_to_math(source):
    r"""
    Replace dollar signs with backticks.

    More precisely, do a regular expression search.  Replace a plain
    dollar sign ($) by a backtick (`).  Replace an escaped dollar sign
    (\$) by a dollar sign ($).  Don't change a dollar sign preceded or
    followed by a backtick (`$ or $`), because of strings like
    "``$HOME``".  Don't make any changes on lines starting with
    spaces, because those are indented and hence part of a block of
    code or examples.

    This also doesn't replaces dollar signs enclosed in curly braces,
    to avoid nested math environments, such as ::

      $f(n) = 0 \text{ if $n$ is prime}$

    Thus the above line would get changed to

      `f(n) = 0 \text{ if $n$ is prime}`
    """
    s = "\n".join(source)
    if s.find("$") == -1:
        return
    # This searches for "$blah$" inside a pair of curly braces --
    # don't change these, since they're probably coming from a nested
    # math environment.  So for each match, we replace it with a temporary
    # string, and later on we substitute the original back.
    _data = {}

    def repl(matchobj):
        nonlocal _data
        s = matchobj.group(0)
        t = f"___XXX_REPL_{len(_data)}___"
        _data[t] = s
        return t

    s = re.sub(r"({[^{}$]*\$[^{}$]*\$[^{}]*})", repl, s)
    # matches $...$
    dollars = re.compile(r"(?<!\$)(?<!\\)\$([^\$]+?)\$")
    # regular expression for \$
    slashdollar = re.compile(r"\\\$")
    s = dollars.sub(r":math:`\1`", s)
    s = slashdollar.sub(r"$", s)
    # change the original {...} things in:
    for r in _data:
        s = s.replace(r, _data[r])
    # now save results in "source"
    source[:] = [s]


def process_dollars(app, docname, source):
    dollars_to_math(source)


def mathdollar_docstrings(app, what, name, obj, options, lines):
    dollars_to_math(lines)


def setup(app):
    app.connect("source-read", process_dollars)
    app.connect('autodoc-process-docstring', mathdollar_docstrings)
    metadata = {"parallel_read_safe": True,
                "parallel_write_safe": True,
                }
    return metadata



import warnings

import numpy.testing as npt
import pytest

from dipy.testing import check_for_warnings
from dipy.testing.decorators import use_xvfb
from dipy.utils.optpkg import optional_package
from dipy.viz.horizon.tab.base import (
    build_checkbox,
    build_label,
    build_radio_button,
    build_slider,
    build_switcher,
)

fury, has_fury, setup_module = optional_package("fury", min_version="0.10.0")

skip_it = use_xvfb == "skip"


def check_label(label):
    npt.assert_equal(label.font_family, "Arial")
    npt.assert_equal(label.justification, "left")
    npt.assert_equal(label.italic, False)
    npt.assert_equal(label.shadow, False)
    npt.assert_equal(
        label.actor.GetTextProperty().GetBackgroundColor(), (0.0, 0.0, 0.0)
    )
    npt.assert_equal(label.actor.GetTextProperty().GetBackgroundOpacity(), 0.0)
    npt.assert_equal(label.color, (0.7, 0.7, 0.7))


@pytest.mark.skipif(skip_it or not has_fury, reason="Needs xvfb")
def test_build_label():
    horizon_label = build_label(text="Hello")
    npt.assert_equal(horizon_label.obj.message, "Hello")
    npt.assert_equal(horizon_label.obj.font_size, 16)
    npt.assert_equal(horizon_label.obj.bold, False)
    check_label(horizon_label.obj)


@pytest.mark.skipif(skip_it or not has_fury, reason="Needs xvfb")
def test_build_slider():
    single_slider_label, single_slider = build_slider(5, 100)
    npt.assert_equal(single_slider_label.obj.message, "")
    npt.assert_equal(single_slider_label.obj.font_size, 16)
    npt.assert_equal(single_slider_label.obj.bold, False)
    npt.assert_equal(single_slider.obj.value, 5)
    npt.assert_equal(single_slider.obj.max_value, 100)
    npt.assert_equal(single_slider.obj.min_value, 0)
    npt.assert_equal(single_slider.obj.track.width, 450)
    npt.assert_equal(single_slider.obj.track.height, 3)
    npt.assert_equal(single_slider.obj.track.color, (0.8, 0.3, 0.0))
    npt.assert_equal(single_slider.obj.default_color, (1.0, 0.5, 0.0))
    npt.assert_equal(single_slider.obj.active_color, (0.9, 0.4, 0.0))
    npt.assert_equal(single_slider.obj.handle.color, (1.0, 0.5, 0.0))
    npt.assert_equal(single_slider.obj.handle.outer_radius, 8)
    npt.assert_equal(single_slider.obj.text.font_size, 16)
    npt.assert_equal(single_slider.obj.text_template, "{value:.1f} ({ratio:.0%})")
    npt.assert_equal(single_slider.selected_value, 5)

    double_slider_label, double_slider = build_slider(
        (4, 5), 100, text_template="{value:.1f}", is_double_slider=True
    )
    npt.assert_equal(double_slider_label.obj.message, "")
    npt.assert_equal(double_slider_label.obj.font_size, 16)
    npt.assert_equal(double_slider_label.obj.bold, False)
    npt.assert_equal(double_slider.obj.max_value, 100)
    npt.assert_equal(double_slider.obj.min_value, 0)
    npt.assert_equal(double_slider.obj.track.width, 450)
    npt.assert_equal(double_slider.obj.track.height, 3)
    npt.assert_equal(double_slider.obj.track.color, (0.8, 0.3, 0.0))
    npt.assert_equal(double_slider.obj.default_color, (1.0, 0.5, 0.0))
    npt.assert_equal(double_slider.obj.active_color, (0.9, 0.4, 0.0))
    npt.assert_equal(double_slider.obj.handles[0].color, (1.0, 0.5, 0.0))
    npt.assert_equal(double_slider.obj.handles[1].color, (1.0, 0.5, 0.0))
    npt.assert_equal(double_slider.obj.handles[0].outer_radius, 8)
    npt.assert_equal(double_slider.obj.handles[1].outer_radius, 8)
    npt.assert_equal(double_slider.obj.text[0].font_size, 16)
    npt.assert_equal(double_slider.obj.text[1].font_size, 16)
    npt.assert_equal(double_slider.selected_value, (4, 5))

    def on_handle_released(a, b, c, d):
        pass

    single_slider_label, single_slider = build_slider(
        5, 100, on_handle_released=on_handle_released
    )
    npt.assert_equal(single_slider_label.obj.message, "")
    npt.assert_equal(single_slider_label.obj.font_size, 16)
    npt.assert_equal(single_slider_label.obj.bold, False)
    npt.assert_equal(single_slider.obj.value, 5)
    npt.assert_equal(single_slider.obj.max_value, 100)
    npt.assert_equal(single_slider.obj.min_value, 0)
    npt.assert_equal(single_slider.obj.track.width, 450)
    npt.assert_equal(single_slider.obj.track.height, 3)
    npt.assert_equal(single_slider.obj.track.color, (0.8, 0.3, 0.0))
    npt.assert_equal(single_slider.obj.default_color, (1.0, 0.5, 0.0))
    npt.assert_equal(single_slider.obj.active_color, (0.9, 0.4, 0.0))
    npt.assert_equal(single_slider.obj.handle.color, (1.0, 0.5, 0.0))
    npt.assert_equal(single_slider.obj.handle.outer_radius, 8)
    npt.assert_equal(single_slider.obj.text.font_size, 16)
    npt.assert_equal(single_slider.obj.text_template, "{value:.1f} ({ratio:.0%})")
    npt.assert_equal(single_slider.selected_value, 5)
    npt.assert_equal(
        single_slider.obj.on_left_mouse_button_released, on_handle_released
    )


@pytest.mark.skipif(skip_it or not has_fury, reason="Needs xvfb")
def test_build_checkbox():
    checkbox = build_checkbox(labels=["Hello", "Hi"], checked_labels=["Hello"])

    npt.assert_equal(len(checkbox.obj.checked_labels), 1)
    npt.assert_equal(len(checkbox.obj.labels), 2)
    npt.assert_equal(checkbox.selected_value, ["Hello"])

    with warnings.catch_warnings(record=True) as l_warns:
        checkbox = build_checkbox()
        npt.assert_equal(checkbox, None)
        check_for_warnings(
            l_warns, "At least one label needs to be to create checkboxes"
        )

    with warnings.catch_warnings(record=True) as l_warns:
        checkbox = build_checkbox([])
        npt.assert_equal(checkbox, None)
        check_for_warnings(
            l_warns, "At least one label needs to be to create checkboxes"
        )


@pytest.mark.skipif(skip_it or not has_fury, reason="Needs xvfb")
def test_build_radio():
    radio = build_radio_button(labels=["Hello", "Hi"], checked_labels=["Hello"])

    npt.assert_equal(len(radio.obj.checked_labels), 1)
    npt.assert_equal(len(radio.obj.labels), 2)
    npt.assert_equal(radio.selected_value, ["Hello"])

    with warnings.catch_warnings(record=True) as l_warns:
        radio = build_radio_button()
        npt.assert_equal(radio, None)
        check_for_warnings(
            l_warns, "At least one label needs to be to create radio buttons"
        )

    with warnings.catch_warnings(record=True) as l_warns:
        radio = build_radio_button([])
        npt.assert_equal(radio, None)
        check_for_warnings(
            l_warns, "At least one label needs to be to create radio buttons"
        )


@pytest.mark.skipif(skip_it or not has_fury, reason="Needs xvfb")
def test_build_switcher():
    with warnings.catch_warnings(record=True) as l_warns:
        none_switcher = build_switcher()
        npt.assert_equal(none_switcher, None)
        check_for_warnings(l_warns, "No items passed in switcher")

    switcher_label, switcher = build_switcher(
        items=[{"label": "Hello", "value": "hello"}]
    )
    npt.assert_equal(switcher.selected_value[0], 0)
    npt.assert_equal(switcher.selected_value[1], "hello")
    npt.assert_equal(switcher_label.selected_value, "")

    switcher_label, switcher = build_switcher(
        items=[{"label": "Hello", "value": "hello"}],
        label="Greeting",
        initial_selection=1,
    )
    npt.assert_equal(switcher.selected_value[0], 0)
    npt.assert_equal(switcher.selected_value[1], "hello")
    npt.assert_equal(switcher_label.selected_value, "Greeting")



import importlib
import os
import sys
import warnings

from numpy.testing import assert_, assert_equal
import pytest

from dipy.utils.optpkg import optional_package

tf, have_tf, _ = optional_package("tensorflow", min_version="2.18.0")
original_backend = os.environ.get("DIPY_NN_BACKEND")
model_mod = None


def setup_module(module):
    """Set up environment variable for all tests in this module."""
    global model_mod
    os.environ["DIPY_NN_BACKEND"] = "tensorflow"
    with warnings.catch_warnings():
        msg = ".*uses TensorFlow.*install PyTorch.*"
        warnings.filterwarnings("ignore", message=msg, category=DeprecationWarning)
        dipy_nn = importlib.reload(sys.modules["dipy.nn"])
    model_mod = dipy_nn.model


def teardown_module(module):
    """Restore the original environment variable after all tests in this module."""
    global model_mod
    if original_backend is not None:
        os.environ["DIPY_NN_BACKEND"] = original_backend
    else:
        del os.environ["DIPY_NN_BACKEND"]
    model_mod = None


@pytest.mark.skipif(not have_tf, reason="Requires TensorFlow")
def test_default_mnist_sequential():
    mnist = tf.keras.datasets.mnist

    epochs = 5

    (x_train, y_train), (x_test, y_test) = mnist.load_data()
    x_train, x_test = x_train / 255.0, x_test / 255.0

    model = tf.keras.models.Sequential(
        [
            tf.keras.layers.Input(shape=(28, 28)),
            tf.keras.layers.Flatten(),
            tf.keras.layers.Dense(128, activation="relu"),
            tf.keras.layers.Dropout(0.2),
            tf.keras.layers.Dense(10, activation="softmax"),
        ]
    )

    model.compile(
        optimizer="adam", loss="sparse_categorical_crossentropy", metrics=["accuracy"]
    )

    hist = model.fit(x_train, y_train, epochs=epochs)
    model.evaluate(x_test, y_test, verbose=2)
    accuracy = hist.history["accuracy"][0]
    assert_(accuracy > 0.9)


@pytest.mark.skipif(not have_tf, reason="Requires TensorFlow")
def test_default_mnist_slp():
    mnist = tf.keras.datasets.mnist
    epochs = 5

    (x_train, y_train), (x_test, y_test) = mnist.load_data()
    x_train, x_test = x_train / 255.0, x_test / 255.0

    slp = model_mod.SingleLayerPerceptron(input_shape=(28, 28))
    hist = slp.fit(x_train, y_train, epochs=epochs)
    slp.evaluate(x_test, y_test, verbose=2)
    x_test_prob = slp.predict(x_test)

    accuracy = hist.history["accuracy"][0]
    assert_(slp.accuracy > 0.9)
    assert_(slp.loss < 0.4)
    assert_equal(slp.accuracy, accuracy)
    assert_equal(x_test_prob.shape, (10000, 10))


@pytest.mark.skipif(not have_tf, reason="Requires TensorFlow")
def test_default_mnist_mlp():
    mnist = tf.keras.datasets.mnist
    epochs = 5

    (x_train, y_train), (x_test, y_test) = mnist.load_data()
    x_train, x_test = x_train / 255.0, x_test / 255.0

    mlp = model_mod.MultipleLayerPercepton(input_shape=(28, 28), num_hidden=[128, 128])
    hist = mlp.fit(x_train, y_train, epochs=epochs)
    mlp.evaluate(x_test, y_test, verbose=2)
    x_test_prob = mlp.predict(x_test)

    accuracy = hist.history["accuracy"][0]
    assert_(mlp.accuracy > 0.8)
    assert_(mlp.loss < 0.4)
    assert_equal(mlp.accuracy, accuracy)
    assert_equal(x_test_prob.shape, (10000, 10))



"""Testing file unit test."""

import sys
import warnings

import numpy as np
import numpy.testing as npt

import dipy.testing as dt


def test_assert():
    npt.assert_raises(AssertionError, dt.assert_false, True)
    npt.assert_raises(AssertionError, dt.assert_true, False)
    npt.assert_raises(AssertionError, dt.assert_less, 2, 1)
    npt.assert_raises(AssertionError, dt.assert_less_equal, 2, 1)
    npt.assert_raises(AssertionError, dt.assert_greater, 1, 2)
    npt.assert_raises(AssertionError, dt.assert_greater_equal, 1, 2)
    npt.assert_raises(AssertionError, dt.assert_not_equal, 5, 5)
    npt.assert_raises(AssertionError, dt.assert_operator, 2, 1)

    arr = [np.arange(k) for k in range(2, 12, 3)]
    arr2 = [np.arange(k) for k in range(2, 12, 4)]
    npt.assert_raises(AssertionError, dt.assert_arrays_equal, arr, arr2)


def assert_warn_len_equal(mod, n_in_context):
    mod_warns = mod.__warningregistry__
    # Python 3 appears to clear any pre-existing warnings of the same type,
    # when raising warnings inside a catch_warnings block. So, there is a
    # warning generated by the tests within the context manager, but no
    # previous warnings.
    if "version" in mod_warns:
        # including 'version'
        npt.assert_equal(len(mod_warns), 2)
    else:
        npt.assert_equal(len(mod_warns), n_in_context)


def test_clear_and_catch_warnings():
    warnings.simplefilter("default", category=UserWarning)
    # Initial state of module, no warnings
    my_mod = sys.modules[__name__]
    try:
        my_mod.__warningregistry__.clear()
    except AttributeError:
        pass

    npt.assert_equal(getattr(my_mod, "__warningregistry__", {}), {})
    with dt.clear_and_catch_warnings(modules=[my_mod]):
        warnings.simplefilter("ignore")
        warnings.warn("Some warning", stacklevel=1)
    npt.assert_equal(my_mod.__warningregistry__, {})
    # Without specified modules, don't clear warnings during context
    with dt.clear_and_catch_warnings():
        warnings.warn("Some warning", stacklevel=1)
    assert_warn_len_equal(my_mod, 1)
    # Confirm that specifying module keeps old warning, does not add new
    with dt.clear_and_catch_warnings(modules=[my_mod]):
        warnings.warn("Another warning", stacklevel=1)
    assert_warn_len_equal(my_mod, 1)
    # Another warning, no module spec does add to warnings dict, except on
    # Python 3 (see comments in `assert_warn_len_equal`)
    with dt.clear_and_catch_warnings():
        warnings.warn("Another warning", stacklevel=1)
    assert_warn_len_equal(my_mod, 2)
    warnings.simplefilter("always", category=UserWarning)



"""
===============================
DSI Deconvolution (DSID) vs DSI
===============================

An alternative method to DSI is the method proposed by
:footcite:p:`CanalesRodriguez2010` which is called DSI with Deconvolution. This
algorithm is using Lucy-Richardson deconvolution in the diffusion propagator
with the goal to create sharper ODFs with higher angular resolution.

In this example we will show with simulated data how this method's ODF
performs against standard DSI ODF and a ground truth multi tensor ODF.
"""

import numpy as np

from dipy.core.gradients import gradient_table
from dipy.data import get_fnames, get_sphere
from dipy.reconst.dsi import DiffusionSpectrumDeconvModel, DiffusionSpectrumModel
from dipy.sims.voxel import multi_tensor, multi_tensor_odf
from dipy.viz import actor, window

###############################################################################
# For the simulation we will use a standard DSI acquisition scheme with 514
# gradient directions and 1 S0.

btable = np.loadtxt(get_fnames(name="dsi515btable"))

gtab = gradient_table(btable[:, 0], bvecs=btable[:, 1:])

###############################################################################
# Let's create a multi tensor with 2 fiber directions at 60 degrees.

evals = np.array([[0.0015, 0.0003, 0.0003], [0.0015, 0.0003, 0.0003]])

directions = [(-30, 0), (30, 0)]

fractions = [50, 50]

signal, _ = multi_tensor(
    gtab, evals, S0=100, angles=directions, fractions=fractions, snr=None
)

sphere = get_sphere(name="repulsion724").subdivide(n=1)

odf_gt = multi_tensor_odf(
    sphere.vertices, evals, angles=directions, fractions=fractions
)

###############################################################################
# Perform the reconstructions with standard DSI and DSI with deconvolution.

dsi_model = DiffusionSpectrumModel(gtab)

dsi_odf = dsi_model.fit(signal).odf(sphere)

dsid_model = DiffusionSpectrumDeconvModel(gtab)

dsid_odf = dsid_model.fit(signal).odf(sphere)

###############################################################################
# Finally, we can visualize the ground truth ODF, together with the DSI and DSI
# with deconvolution ODFs and observe that with the deconvolved method it is
# easier to resolve the correct fiber directions because the ODF is sharper.

# Enables/disables interactive visualization
interactive = False


scene = window.Scene()

# concatenate data as 4D array
odfs = np.vstack((odf_gt, dsi_odf, dsid_odf))[:, None, None]
odf_actor = actor.odf_slicer(odfs, sphere=sphere, scale=0.5, colormap="plasma")

odf_actor.display(y=0)
odf_actor.RotateX(90)
scene.add(odf_actor)
window.record(scene=scene, out_path="dsid.png", size=(300, 300))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# Ground truth ODF (left), DSI ODF (middle), DSI with Deconvolution ODF (right).
#
#
# References
# ----------
#
# .. footbibliography::
#



import numpy as np

from dipy.viz.horizon.tab import HorizonTab, build_slider


class ClustersTab(HorizonTab):
    def __init__(self, clusters_visualizer, threshold):
        """Initialize Interaction tab for cluster visualization.

        Parameters
        ----------
        clusters_visualizer : ClusterVisualizer
        threshold : float
        """

        super().__init__()

        self._visualizer = clusters_visualizer

        self._name = "Clusters"

        self._tab_id = 0

        sizes = self._visualizer.sizes
        self._size_slider_label, self._size_slider = build_slider(
            initial_value=np.percentile(sizes, 50),
            min_value=np.min(sizes),
            max_value=np.percentile(sizes, 98),
            text_template="{value:.0f}",
            label="Size",
            on_change=self._change_size,
        )

        lengths = self._visualizer.lengths
        self._length_slider_label, self._length_slider = build_slider(
            initial_value=np.percentile(lengths, 25),
            min_value=np.min(lengths),
            max_value=np.percentile(lengths, 98),
            text_template="{value:.0f}",
            label="Length",
            on_change=self._change_length,
        )

        self._threshold_slider_label, self._threshold_slider = build_slider(
            initial_value=threshold,
            min_value=5,
            max_value=25,
            text_template="{value:.0f}",
            label="Threshold",
            on_handle_released=self._change_threshold,
        )

        self._register_elements(
            self._size_slider_label,
            self._size_slider,
            self._length_slider_label,
            self._length_slider,
            self._threshold_slider_label,
            self._threshold_slider,
        )

    def _change_length(self, slider):
        """Change the length threshold for visibility.

        Parameters
        ----------
        slider : LineSlider2D
            FURY object for slider.
        """
        self._length_slider.selected_value = int(np.rint(slider.value))
        self._update_clusters()

    def _change_size(self, slider):
        """Change the size threshold for visibility.

        Parameters
        ----------
        slider : LineSlider2D
            FURY object for slider.
        """
        self._size_slider.selected_value = int(np.rint(slider.value))
        self._update_clusters()

    def _change_threshold(self, _istyle, _obj, slider):
        """Re-cluster the tractograms according to the new threshold set. It
        will also update the size and length slider for corresponding
        threshold.

        Parameters
        ----------
        _istyle : vtkInteractor
            Should not be used.
        _obj : vtkObject
            Should not be used.
        slider : LineSlider2D
            FURY object for slider.
        """
        value = int(np.rint(slider.value))
        if value != self._threshold_slider.selected_value:
            self._visualizer.recluster_tractograms(value)

            sizes = self._visualizer.sizes
            self._size_slider.selected_value = np.percentile(sizes, 50)

            lengths = self._visualizer.lengths
            self._length_slider.selected_value = np.percentile(lengths, 25)

            self._update_clusters()

            self._size_slider.obj.min_value = np.min(sizes)
            self._size_slider.obj.max_value = np.percentile(sizes, 98)
            self._size_slider.obj.value = self._size_slider.selected_value
            self._size_slider.obj.update()

            self._length_slider.obj.min_value = np.min(lengths)
            self._length_slider.obj.max_value = np.percentile(lengths, 98)
            self._length_slider.obj.value = self._length_slider.selected_value
            self._length_slider.obj.update()

            self._threshold_slider.selected_value = value

    def _update_clusters(self):
        """Updates the clusters according to set size and length."""
        for k, cluster in self.cluster_actors.items():
            length_validation = cluster["length"] < self._length_slider.selected_value

            size_validation = cluster["size"] < self._size_slider.selected_value

            if length_validation or size_validation:
                cluster["actor"].SetVisibility(False)

                if k.GetVisibility():
                    k.SetVisibility(False)
            else:
                cluster["actor"].SetVisibility(True)

    def build(self, tab_id):
        """Position the elements in the tab.

        Parameters
        ----------
        tab_id : int
            Id of the tab.
        """
        self._tab_id = tab_id

        x_pos = 0.02

        self._size_slider_label.position = (x_pos, 0.85)
        self._length_slider_label.position = (x_pos, 0.62)
        self._threshold_slider_label.position = (x_pos, 0.38)

        x_pos = 0.12

        self._size_slider.position = (x_pos, 0.85)
        self._length_slider.position = (x_pos, 0.62)
        self._threshold_slider.position = (x_pos, 0.38)

    @property
    def name(self):
        """Title of the tab.

        Returns
        -------
        str
        """
        return self._name

    @property
    def cluster_actors(self):
        """Cluster actors of the tractograms.

        Returns
        -------
        dict
            various properties of clusters.
        """
        return self._visualizer.cluster_actors

    @property
    def centroid_actors(self):
        """Centroid actors of the tractograms.

        Returns
        -------
        dict
            various properties of centroids.
        """
        return self._visualizer.centroid_actors

    @property
    def actors(self):
        """All the actors in the visualizer.

        Returns
        -------
        list
        """
        actors = []
        for cluster_actor in self.cluster_actors.values():
            actors.append(cluster_actor["actor"])
        for centroid_actor in self.centroid_actors.values():
            actors.append(centroid_actor["actor"])
        return actors



"""
====================================
BUAN Bundle Assignment Maps Creation
====================================

This example explains how we can use BUAN :footcite:p:`Chandio2020a` to create
assignment maps on a bundle. Divide bundle into N smaller segments.


First import the necessary modules.
"""

import numpy as np

from dipy.data import fetch_bundle_atlas_hcp842, get_two_hcp842_bundles
from dipy.io.streamline import load_trk
from dipy.stats.analysis import assignment_map
from dipy.viz import actor, window

###############################################################################
# Download and read data for this tutorial

atlas_file, atlas_folder = fetch_bundle_atlas_hcp842()

###############################################################################
# Read AF left and CST left bundles from already fetched atlas data to use them
# as model bundles

model_af_l_file, model_cst_l_file = get_two_hcp842_bundles()

sft_af_l = load_trk(model_af_l_file, reference="same", bbox_valid_check=False)
model_af_l = sft_af_l.streamlines

###############################################################################
# let's visualize Arcuate Fasiculus Left (AF_L) bundle before assignment maps

interactive = False

scene = window.Scene()
scene.SetBackground(1, 1, 1)
scene.add(actor.line(model_af_l, fake_tube=True, linewidth=6))
scene.set_camera(
    focal_point=(-18.17281532, -19.55606842, 6.92485857),
    position=(-360.11, -30.46, -40.44),
    view_up=(-0.03, 0.028, 0.89),
)
window.record(scene=scene, out_path="af_l_before_assignment_maps.png", size=(600, 600))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# AF_L before assignment maps
#
#
#
# Creating 100 bundle assignment maps on AF_L using BUAN
# :footcite:p:`Chandio2020a`

rng = np.random.default_rng()

n = 100
indx = assignment_map(model_af_l, model_af_l, n)
indx = np.array(indx)

colors = [rng.random(3) for si in range(n)]

disks_color = []
for i in range(len(indx)):
    disks_color.append(tuple(colors[indx[i]]))

###############################################################################
# let's visualize Arcuate Fasiculus Left (AF_L) bundle after assignment maps

interactive = False

scene = window.Scene()
scene.SetBackground(1, 1, 1)
scene.add(actor.line(model_af_l, fake_tube=True, colors=disks_color, linewidth=6))
scene.set_camera(
    focal_point=(-18.17281532, -19.55606842, 6.92485857),
    position=(-360.11, -30.46, -40.44),
    view_up=(-0.03, 0.028, 0.89),
)
window.record(scene=scene, out_path="af_l_after_assignment_maps.png", size=(600, 600))
if interactive:
    window.show(scene)

###############################################################################
# .. rst-class:: centered small fst-italic fw-semibold
#
# AF_L after assignment maps
#
#
# References
# ----------
#
# .. footbibliography::
#



from glob import glob
import json
import logging
import os
from time import time
import warnings

import numpy as np
from scipy.ndimage import binary_dilation

from dipy.core.gradients import gradient_table
from dipy.io import read_bvals_bvecs
from dipy.io.image import load_nifti, save_nifti
from dipy.io.peaks import load_peaks
from dipy.io.streamline import load_tractogram
from dipy.reconst.dti import TensorModel
from dipy.segment.bundles import bundle_shape_similarity
from dipy.segment.mask import bounding_box, segment_from_cfa
from dipy.stats.analysis import anatomical_measures, assignment_map, peak_values
from dipy.testing.decorators import warning_for_keywords
from dipy.tracking.streamline import transform_streamlines
from dipy.utils.optpkg import optional_package
from dipy.workflows.workflow import Workflow

pd, have_pd, _ = optional_package("pandas")
smf, have_smf, _ = optional_package("statsmodels.formula.api")
matplt, have_matplotlib, _ = optional_package("matplotlib")
plt, have_plt, _ = optional_package("matplotlib.pyplot")


class SNRinCCFlow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "snrincc"

    def run(
        self,
        data_files,
        bvals_files,
        bvecs_files,
        mask_file,
        bbox_threshold=(0.6, 1, 0, 0.1, 0, 0.1),
        out_dir="",
        out_file="product.json",
        out_mask_cc="cc.nii.gz",
        out_mask_noise="mask_noise.nii.gz",
    ):
        """Compute the signal-to-noise ratio in the corpus callosum.

        Parameters
        ----------
        data_files : string
            Path to the dwi.nii.gz file. This path may contain wildcards to
            process multiple inputs at once.
        bvals_files : string
            Path of bvals.
        bvecs_files : string
            Path of bvecs.
        mask_file : string
            Path of a brain mask file.
        bbox_threshold : variable float, optional
            Threshold for bounding box, values separated with commas for ex.
            [0.6,1,0,0.1,0,0.1].
        out_dir : string, optional
            Where the resulting file will be saved.
        out_file : string, optional
            Name of the result file to be saved.
        out_mask_cc : string, optional
            Name of the CC mask volume to be saved.
        out_mask_noise : string, optional
            Name of the mask noise volume to be saved.

        """
        io_it = self.get_io_iterator()

        for (
            dwi_path,
            bvals_path,
            bvecs_path,
            mask_path,
            out_path,
            cc_mask_path,
            mask_noise_path,
        ) in io_it:
            data, affine = load_nifti(dwi_path)
            bvals, bvecs = read_bvals_bvecs(bvals_path, bvecs_path)
            gtab = gradient_table(bvals=bvals, bvecs=bvecs)

            mask, affine = load_nifti(mask_path)

            logging.info("Computing tensors...")
            tenmodel = TensorModel(gtab)
            tensorfit = tenmodel.fit(data, mask=mask)

            logging.info("Computing worst-case/best-case SNR using the CC...")

            if np.ndim(data) == 4:
                CC_box = np.zeros_like(data[..., 0])
            elif np.ndim(data) == 3:
                CC_box = np.zeros_like(data)
            else:
                raise OSError("DWI data has invalid dimensions")

            mins, maxs = bounding_box(mask)
            mins = np.array(mins)
            maxs = np.array(maxs)
            diff = (maxs - mins) // 4
            bounds_min = mins + diff
            bounds_max = maxs - diff

            CC_box[
                bounds_min[0] : bounds_max[0],
                bounds_min[1] : bounds_max[1],
                bounds_min[2] : bounds_max[2],
            ] = 1

            if len(bbox_threshold) != 6:
                raise OSError("bbox_threshold should have 6 float values")

            mask_cc_part, cfa = segment_from_cfa(
                tensorfit, CC_box, bbox_threshold, return_cfa=True
            )

            if not np.count_nonzero(mask_cc_part.astype(np.uint8)):
                logging.warning(
                    "Empty mask: corpus callosum not found."
                    " Update your data or your threshold"
                )

            save_nifti(cc_mask_path, mask_cc_part.astype(np.uint8), affine)
            logging.info(f"CC mask saved as {cc_mask_path}")

            masked_data = data[mask_cc_part]
            mean_signal = 0
            if masked_data.size:
                mean_signal = np.mean(masked_data, axis=0)
            mask_noise = binary_dilation(mask, iterations=10)
            mask_noise[..., : mask_noise.shape[-1] // 2] = 1
            mask_noise = ~mask_noise

            save_nifti(mask_noise_path, mask_noise.astype(np.uint8), affine)
            logging.info(f"Mask noise saved as {mask_noise_path}")

            noise_std = 0
            if np.count_nonzero(mask_noise.astype(np.uint8)):
                noise_std = np.std(data[mask_noise, :])

            logging.info(f"Noise standard deviation sigma= {noise_std}")

            idx = np.sum(gtab.bvecs, axis=-1) == 0
            gtab.bvecs[idx] = np.inf
            axis_X = np.argmin(np.sum((gtab.bvecs - np.array([1, 0, 0])) ** 2, axis=-1))
            axis_Y = np.argmin(np.sum((gtab.bvecs - np.array([0, 1, 0])) ** 2, axis=-1))
            axis_Z = np.argmin(np.sum((gtab.bvecs - np.array([0, 0, 1])) ** 2, axis=-1))

            SNR_output = []
            SNR_directions = []
            for direction in ["b0", axis_X, axis_Y, axis_Z]:
                if direction == "b0":
                    SNR = mean_signal[0] / noise_std if noise_std else 0
                    logging.info(f"SNR for the b=0 image is : {SNR}")
                else:
                    logging.info(
                        f"SNR for direction {direction} {gtab.bvecs[direction]} is: "
                        f"{SNR}"
                    )
                    SNR_directions.append(direction)
                    SNR = mean_signal[direction] / noise_std if noise_std else 0
                SNR_output.append(SNR)

            snr_str = f"{SNR_output[0]} {SNR_output[1]} {SNR_output[2]} {SNR_output[3]}"
            dir_str = f"b0 {SNR_directions[0]} {SNR_directions[1]} {SNR_directions[2]}"
            data = [{"data": snr_str, "directions": dir_str}]

            with open(os.path.join(out_dir, out_path), "w") as myfile:
                json.dump(data, myfile)


@warning_for_keywords()
def buan_bundle_profiles(
    model_bundle_folder,
    bundle_folder,
    orig_bundle_folder,
    metric_folder,
    group_id,
    subject,
    *,
    no_disks=100,
    out_dir="",
):
    """
    Applies statistical analysis on bundles and saves the results
    in a directory specified by ``out_dir``.

    See :footcite:p:`Chandio2020a` for further details about the method.

    Parameters
    ----------
    model_bundle_folder : string
        Path to the input model bundle files. This path may contain
        wildcards to process multiple inputs at once.
    bundle_folder : string
        Path to the input bundle files in common space. This path may
        contain wildcards to process multiple inputs at once.
    orig_bundle_folder : string
        Path to the input bundle files in native space. This path may
        contain wildcards to process multiple inputs at once.
    metric_folder : string
        Path to the input dti metric or/and peak files. It will be used as
        metric for statistical analysis of bundles.
    group_id : integer
        what group subject belongs to either 0 for control or 1 for patient.
    subject : string
        subject id e.g. 10001.
    no_disks : integer, optional
        Number of disks used for dividing bundle into disks.
    out_dir : string, optional
        Output directory.

    References
    ----------
    .. footbibliography::

    """

    t = time()

    mb = glob(os.path.join(model_bundle_folder, "*.trk"))
    print(mb)

    mb.sort()

    bd = glob(os.path.join(bundle_folder, "*.trk"))

    bd.sort()
    print(bd)
    org_bd = glob(os.path.join(orig_bundle_folder, "*.trk"))
    org_bd.sort()
    print(org_bd)
    n = len(mb)

    for io in range(n):
        mbundles = load_tractogram(
            mb[io], reference="same", bbox_valid_check=False
        ).streamlines
        bundles = load_tractogram(
            bd[io], reference="same", bbox_valid_check=False
        ).streamlines
        orig_bundles = load_tractogram(
            org_bd[io], reference="same", bbox_valid_check=False
        ).streamlines

        if len(orig_bundles) > 5:
            indx = assignment_map(bundles, mbundles, no_disks)
            ind = np.array(indx)

            metric_files_names_dti = glob(os.path.join(metric_folder, "*.nii.gz"))

            metric_files_names_csa = glob(os.path.join(metric_folder, "*.pam5"))

            _, affine = load_nifti(metric_files_names_dti[0])

            affine_r = np.linalg.inv(affine)
            transformed_orig_bundles = transform_streamlines(orig_bundles, affine_r)

            for mn in range(len(metric_files_names_dti)):
                ab = os.path.split(metric_files_names_dti[mn])
                metric_name = ab[1]

                fm = metric_name[:-7]
                bm = os.path.split(mb[io])[1][:-4]

                logging.info(f"bm = {bm}")

                dt = {}

                logging.info(f"metric = {metric_files_names_dti[mn]}")

                metric, _ = load_nifti(metric_files_names_dti[mn])

                anatomical_measures(
                    transformed_orig_bundles,
                    metric,
                    dt,
                    fm,
                    bm,
                    subject,
                    group_id,
                    ind,
                    out_dir,
                )

            for mn in range(len(metric_files_names_csa)):
                ab = os.path.split(metric_files_names_csa[mn])
                metric_name = ab[1]

                fm = metric_name[:-5]
                bm = os.path.split(mb[io])[1][:-4]

                logging.info(f"bm = {bm}")
                logging.info(f"metric = {metric_files_names_csa[mn]}")
                dt = {}
                metric = load_peaks(metric_files_names_csa[mn])

                peak_values(
                    transformed_orig_bundles,
                    metric,
                    dt,
                    fm,
                    bm,
                    subject,
                    group_id,
                    ind,
                    out_dir,
                )

    print("total time taken in minutes = ", (-t + time()) / 60)


class BundleAnalysisTractometryFlow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "ba"

    @warning_for_keywords()
    def run(self, model_bundle_folder, subject_folder, *, no_disks=100, out_dir=""):
        """Workflow of bundle analytics.

        Applies statistical analysis on bundles of subjects and saves the
        results in a directory specified by ``out_dir``.

        See :footcite:p:`Chandio2020a` for further details about the method.

        Parameters
        ----------
        model_bundle_folder : string
            Path to the input model bundle files. This path may
            contain wildcards to process multiple inputs at once.
        subject_folder : string
            Path to the input subject folder. This path may contain
            wildcards to process multiple inputs at once.
        no_disks : integer, optional
            Number of disks used for dividing bundle into disks.
        out_dir : string, optional
            Output directory.

        References
        ----------
        .. footbibliography::

        """

        if os.path.isdir(subject_folder) is False:
            raise ValueError("Invalid path to subjects")

        groups = os.listdir(subject_folder)
        groups.sort()
        for group in groups:
            if os.path.isdir(os.path.join(subject_folder, group)):
                logging.info(f"group = {group}")
                all_subjects = os.listdir(os.path.join(subject_folder, group))
                all_subjects.sort()
                logging.info(all_subjects)
            if group.lower() == "patient":
                group_id = 1  # 1 means patient
            elif group.lower() == "control":
                group_id = 0  # 0 means control
            else:
                print(group)
                raise ValueError("Invalid group. Neither patient nor control")

            for sub in all_subjects:
                logging.info(sub)
                pre = os.path.join(subject_folder, group, sub)
                logging.info(pre)
                b = os.path.join(pre, "rec_bundles")
                c = os.path.join(pre, "org_bundles")
                d = os.path.join(pre, "anatomical_measures")
                buan_bundle_profiles(
                    model_bundle_folder,
                    b,
                    c,
                    d,
                    group_id,
                    sub,
                    no_disks=no_disks,
                    out_dir=out_dir,
                )


class LinearMixedModelsFlow(Workflow):
    @classmethod
    def get_short_name(cls):
        return "lmm"

    def get_metric_name(self, path):
        """Splits the path string and returns name of anatomical measure
        (eg: fa), bundle name eg(AF_L) and bundle name with metric name
        (eg: AF_L_fa)

        Parameters
        ----------
        path : string
            Path to the input metric files. This path may
            contain wildcards to process multiple inputs at once.
        """

        head_tail = os.path.split(path)
        name = head_tail[1]
        count = 0
        i = len(name) - 1
        while i > 0:
            if name[i] == ".":
                count = i
                break
            i = i - 1

        for j in range(len(name)):
            if name[j] == "_":
                if name[j + 1] != "L" and name[j + 1] != "R" and name[j + 1] != "F":
                    return name[j + 1 : count], name[:j], name[:count]

        return " ", " ", " "

    def save_lmm_plot(self, plot_file, title, bundle_name, x, y):
        """Saves LMM plot with segment/disk number on x-axis and
        -log10(pvalues) on y-axis in out_dir folder.

        Parameters
        ----------
        plot_file : string
            Path to the plot file. This path may
            contain wildcards to process multiple inputs at once.
        title : string
            Title for the plot.
        bundle_name : string
            Bundle name.
        x : list
            list containing segment/disk number for x-axis.
        y : list
            list containing -log10(pvalues) per segment/disk number for y-axis.

        """

        n = len(x)
        dotted = np.ones(n)
        dotted[:] = 2
        c1 = np.random.rand(1, 3)

        y_pos = np.arange(n)

        (l1,) = plt.plot(
            y_pos,
            dotted,
            color="red",
            marker=".",
            linestyle="solid",
            linewidth=0.6,
            markersize=0.7,
            label="p-value < 0.01",
        )

        (l2,) = plt.plot(
            y_pos,
            dotted + 1,
            color="black",
            marker=".",
            linestyle="solid",
            linewidth=0.4,
            markersize=0.4,
            label="p-value < 0.001",
        )

        first_legend = plt.legend(handles=[l1, l2], loc="upper right")

        axes = plt.gca()
        axes.add_artist(first_legend)
        axes.set_ylim([0, 6])

        l3 = plt.bar(y_pos, y, color=c1, alpha=0.5, label=bundle_name)
        plt.legend(handles=[l3], loc="upper left")
        plt.title(title.upper())
        plt.xlabel("Segment Number")
        plt.ylabel("-log10(Pvalues)")
        plt.savefig(plot_file)
        plt.clf()

    @warning_for_keywords()
    def run(self, h5_files, *, no_disks=100, out_dir=""):
        """Workflow of linear Mixed Models.

        Applies linear Mixed Models on bundles of subjects and saves the
        results in a directory specified by ``out_dir``.

        Parameters
        ----------
        h5_files : string
            Path to the input metric files. This path may
            contain wildcards to process multiple inputs at once.
        no_disks : integer, optional
            Number of disks used for dividing bundle into disks.
        out_dir : string, optional
            Output directory.

        """

        io_it = self.get_io_iterator()

        for file_path in io_it:
            logging.info(f"Applying metric {file_path}")

            file_name, bundle_name, save_name = self.get_metric_name(file_path)
            logging.info(f" file name = {file_name}")
            logging.info(f"file path = {file_path}")

            pvalues = np.zeros(no_disks)
            warnings.filterwarnings("ignore")
            # run mixed linear model for every disk
            for i in range(no_disks):
                disk_count = i + 1
                df = pd.read_hdf(file_path, where="disk=disk_count")

                logging.info(f"read the dataframe for disk number {disk_count}")
                # check if data has significant data to perform LMM
                if len(df) < 10:
                    raise ValueError("Dataset for Linear Mixed Model is too small")

                criteria = file_name + " ~ group"
                md = smf.mixedlm(criteria, df, groups=df["subject"])

                mdf = md.fit()

                pvalues[i] = mdf.pvalues[1]

            x = list(range(1, len(pvalues) + 1))
            y = -1 * np.log10(pvalues)

            save_file = os.path.join(out_dir, save_name + "_pvalues.npy")
            np.save(save_file, pvalues)

            save_file = os.path.join(out_dir, save_name + "_pvalues_log.npy")
            np.save(save_file, y)

            save_file = os.path.join(out_dir, save_name + ".png")
            self.save_lmm_plot(save_file, file_name, bundle_name, x, y)


class BundleShapeAnalysis(Workflow):
    @classmethod
    def get_short_name(cls):
        return "BS"

    @warning_for_keywords()
    def run(self, subject_folder, *, clust_thr=(5, 3, 1.5), threshold=6, out_dir=""):
        """Workflow of bundle analytics.

        Applies bundle shape similarity analysis on bundles of subjects and
        saves the results in a directory specified by ``out_dir``.

        See :footcite:p:`Chandio2020a` for further details about the method.

        Parameters
        ----------
        subject_folder : string
            Path to the input subject folder. This path may contain
            wildcards to process multiple inputs at once.
        clust_thr : variable float, optional
            list of bundle clustering thresholds used in QuickBundlesX.
        threshold : float, optional
            Bundle shape similarity threshold.
        out_dir : string, optional
            Output directory.

        References
        ----------
        .. footbibliography::

        """
        rng = np.random.default_rng()
        all_subjects = []
        if os.path.isdir(subject_folder):
            groups = os.listdir(subject_folder)
            groups.sort()
        else:
            raise ValueError("Not a directory")

        for group in groups:
            if os.path.isdir(os.path.join(subject_folder, group)):
                subjects = os.listdir(os.path.join(subject_folder, group))
                subjects.sort()
                logging.info(
                    "first "
                    + str(len(subjects))
                    + " subjects in matrix belong to "
                    + group
                    + " group"
                )

                for sub in subjects:
                    dpath = os.path.join(subject_folder, group, sub)
                    if os.path.isdir(dpath):
                        all_subjects.append(dpath)

        N = len(all_subjects)

        bundles = os.listdir(os.path.join(all_subjects[0], "rec_bundles"))
        for bun in bundles:
            # bundle shape similarity matrix
            ba_matrix = np.zeros((N, N))
            i = 0
            logging.info(bun)
            for sub in all_subjects:
                j = 0

                bundle1 = load_tractogram(
                    os.path.join(sub, "rec_bundles", bun),
                    reference="same",
                    bbox_valid_check=False,
                ).streamlines

                for subi in all_subjects:
                    logging.info(subi)

                    bundle2 = load_tractogram(
                        os.path.join(subi, "rec_bundles", bun),
                        reference="same",
                        bbox_valid_check=False,
                    ).streamlines

                    ba_value = bundle_shape_similarity(
                        bundle1, bundle2, rng, clust_thr=clust_thr, threshold=threshold
                    )

                    ba_matrix[i][j] = ba_value

                    j += 1
                i += 1
            logging.info("saving BA score matrix")
            np.save(os.path.join(out_dir, bun[:-4] + ".npy"), ba_matrix)

            cmap = matplt.colormaps["Blues"]
            plt.title(bun[:-4])
            plt.imshow(ba_matrix, cmap=cmap)
            plt.colorbar()
            plt.clim(0, 1)
            plt.savefig(os.path.join(out_dir, f"SM_{bun[:-4]}"))
            plt.clf()
